#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from keystoneauth1 import loading as ks_loading
from keystoneauth1 import service_token
from oslo_log import log as logging

import cyborg.conf


CONF = cyborg.conf.CONF
LOG = logging.getLogger(__name__)

_SERVICE_AUTH = None


def reset_globals():
    """For async unit test consistency."""
    global _SERVICE_AUTH
    _SERVICE_AUTH = None


def get_auth_plugin(context):
    user_auth = context.get_auth_plugin()

    if CONF.service_user.send_service_user_token:
        global _SERVICE_AUTH
        if not _SERVICE_AUTH:
            _SERVICE_AUTH = ks_loading.\
                load_auth_from_conf_options(CONF,
                                            group=cyborg.
                                            conf.service_token.
                                            SERVICE_USER_GROUP)
            if _SERVICE_AUTH is None:
                # This indicates a misconfiguration so log a warning and
                # return the user_auth.
                LOG.warning('Unable to load auth from [service_user] '
                            'configuration. Ensure "auth_type" is set.')
                return user_auth
        return service_token.\
            ServiceTokenAuthWrapper(user_auth=user_auth,
                                    service_auth=_SERVICE_AUTH)

    return user_auth



# Copyright 2019 Intel Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.objects import arq
from cyborg.objects import ext_arq
from cyborg.objects.extarq import fpga_ext_arq


def _get_arqs_as_dict():
    common = {
        "state": "Bound",
        "device_profile_name": "dp_example_1",
        "hostname": "myhost",
        "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
        "attach_handle_type": "PCI",
        # attach_handle info should vary across ARQs but ignored for testing
        "attach_handle_info": {
            "bus": "1",
            "device": "0",
            "domain": "0",
            "function": "0"
        },
        "device_profile_group": {
            "trait:CUSTOM_FPGA_INTEL": "required",
            "resources:FPGA": "1",
            "accel:bitstream_id": "b069d97a-010a-4057-b70d-eca2b337fc9c"}
    }
    dp_groups = [
        {"device_profile_group": {"resources:GPU": "1"}},
        {"device_profile_group": {
            "trait:CUSTOM_FPGA_INTEL": "required",
            "resources:FPGA": "1"}},
        {"device_profile_group": {
            "trait:CUSTOM_FPGA_INTEL": "required",
            "resources:FPGA": "1",
            "accel:bitstream_id": "b069d97a-010a-4057-b70d-eca2b337fc9c"}},
        {"device_profile_group": {
            "trait:CUSTOM_FPGA_INTEL": "required",
            "resources:FPGA": "1",
            "accel:function_id": "25453786-03e0-4ee7-a640-969eb5a5aa44"}},
        {"device_profile_group": {
            "trait:CUSTOM_FPGA_INTEL": "required",
            "resources:FPGA": "1",
            "accel:bitstream_id": "b069d97a-010a-4057-b70d-eca2b337fc9c",
            "accel:function_id": "25453786-03e0-4ee7-a640-969eb5a5aa44"}},
    ]
    arqs = [  # Corresponds to 1st device profile in fake_device)profile.py
        {"uuid": "a097fefa-da62-4630-8e8b-424c0e3426dc",
         "device_profile_group_id": 0,
         "device_rp_uuid": "8787595e-9954-49f8-b5c1-cdb55b59062f",
         },
        {"uuid": "aa140114-4869-45ea-8213-45f530804b0f",
         "device_profile_group_id": 1,
         "device_rp_uuid": "a1ec17f2-0051-4737-bac4-f074d8a01a9c",
         },
        {"uuid": "292b2fa2-0831-484c-aeac-09c794428a5d",
         "device_profile_group_id": 2,
         "device_rp_uuid": "a1ec17f2-0051-4737-bac4-f074d8a01a9c",
         },
        {"uuid": "3049ad04-a2b1-40a3-b9c8-480a5e661645",
         "device_profile_group_id": 3,
         "device_rp_uuid": "57455a49-bde4-490e-9179-9aa84a3870bb",
         },
        {"uuid": "3a9a07e7-d126-47a5-bf11-dcc04f9e60ff",
         "device_profile_group_id": 4,
         "device_rp_uuid": "fbd485e1-40b1-4a7e-84b9-f6b6959114a4",
         },
    ]
    new_arqs = []
    for idx, new_arq in enumerate(arqs):
        common.update(dp_groups[idx])
        new_arq.update(common)
        new_arq.update(id=idx)
        new_arqs.append(new_arq)
    return new_arqs


def _get_arqs_resloved_as_dict():
    arqs = [  # Corresponds to 1st device profile in fake_device)profile.py
        {"uuid": 'a097fefa-da62-4630-8e8b-424c0e3426dd',
         "device_profile_group_id": 0,
         "state": "Initial",
         "device_profile_name": "dp_example_1",
         "device_rp_uuid": None,
         "hostname": None,
         "instance_uuid": None,
         "attach_handle_type": None,
         # attach_handle info should vary across ARQs but ignored for testing
         "attach_handle_info": {},
         "device_profile_group": {}
         },
        {"uuid": 'aa140114-4869-45ea-8213-45f530804b0e',
         "device_profile_group_id": 1,
         "device_rp_uuid": "fbd485e1-40b1-4a7e-84b9-f6b6959114a5",
         "state": "Bound",
         "device_profile_name": "dp_example_1",
         "hostname": "myhost",
         "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
         "attach_handle_type": "PCI",
         # attach_handle info should vary across ARQs but ignored for testing
         "attach_handle_info": {
             "bus": "1",
             "device": "0",
             "domain": "0",
             "function": "0"
         },
         "device_profile_group": {
             "trait:CUSTOM_FPGA_INTEL": "required",
             "resources:FPGA": "1",
             "accel:bitstream_id": "b069d97a-010a-4057-b70d-eca2b337fc9c"}
         },
    ]
    new_arqs = []
    for idx, new_arq in enumerate(arqs):
        new_arq.update(id=idx)
        new_arqs.append(new_arq)
    return new_arqs


def _get_arqs_bind_as_dict():
    common = [
        {
            "state": "Bound",
            "device_profile_name": "dp_example_1",
            "hostname": "myhost1",
            "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
            "attach_handle_type": "PCI",
            # attach_handle info should vary across ARQs but ignored for
            # testing
            "attach_handle_info": {
                "bus": "1",
                "device": "0",
                "domain": "0",
                "function": "0"
            },
            "device_profile_group": {"resources:GPU": "1"}
        }, {
            "state": "Deleting",
            "device_profile_name": "dp_example_2",
            "hostname": "myhost1",
            "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
            "attach_handle_type": "PCI",
            # attach_handle info should vary across ARQs but ignored for
            # testing
            "attach_handle_info": {
                "bus": "2",
                "device": "0",
                "domain": "0",
                "function": "0"
            },
            "device_profile_group": {
                "trait:CUSTOM_FPGA_INTEL": "required",
                "resources:FPGA": "1"}
        }, {
            "state": "Deleting",
            "device_profile_name": "dp_example_3",
            "hostname": "myhost3",
            "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7146",
            "attach_handle_type": "PCI",
            # attach_handle info should vary across ARQs but ignored for
            # testing
            "attach_handle_info": {
                "bus": "3",
                "device": "0",
                "domain": "0",
                "function": "0"
            },
            "device_profile_group": {
                "trait:CUSTOM_FPGA_INTEL": "required",
                "resources:FPGA": "1",
                "accel:bitstream_id": "b069d97a-010a-4057-b70d-eca2b337fc9e"}
        }, {
            "state": "Unbound",
            "device_profile_name": "dp_example_2",
            "hostname": "myhost1",
            "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
            "attach_handle_type": "PCI",
            # attach_handle info should vary across ARQs but ignored for
            # testing
            "attach_handle_info": {
                "bus": "2",
                "device": "0",
                "domain": "0",
                "function": "0"
            },
            "device_profile_group": {
                "trait:CUSTOM_FPGA_INTEL": "required",
                "resources:FPGA": "1"}
        },
    ]
    arqs = [  # Corresponds to 1st device profile in fake_device)profile.py
        {"uuid": "a097fefa-da62-4630-8e8b-424c0e3426de",
         "device_profile_group_id": 0,
         "device_rp_uuid": "8787595e-9954-49f8-b5c1-cdb55b59062e",
         },
        {"uuid": "aa140114-4869-45ea-8213-45f530804b0d",
         "device_profile_group_id": 0,
         "device_rp_uuid": "a1ec17f2-0051-4737-bac4-f074d8a01a9d",
         },
        {"uuid": "292b2fa2-0831-484c-aeac-09c794428a5e",
         "device_profile_group_id": 0,
         "device_rp_uuid": "57455a49-bde4-490e-9179-9aa84a3870bb",
         },
        {"uuid": "292b2fa2-0831-484c-aeac-09c794428a5d",
         "device_profile_group_id": 0,
         "device_rp_uuid": "57455a49-bde4-490e-9179-9aa84a3870bc",
         }
    ]
    new_arqs = []
    for idx, new_arq in enumerate(arqs):
        new_arq.update(common[idx])
        new_arq.update(id=idx)
        new_arqs.append(new_arq)
    return new_arqs


def _convert_from_dict_to_obj(arq_dict):
    obj_arq = arq.ARQ()
    for field in arq_dict.keys():
        obj_arq[field] = arq_dict[field]
    obj_extarq = ext_arq.ExtARQ()
    obj_extarq.arq = obj_arq
    obj_extarq.device_profile_group = arq_dict["device_profile_group"]
    return obj_extarq


def get_fake_extarq_objs():
    arq_list = _get_arqs_as_dict()
    obj_extarqs = list(map(_convert_from_dict_to_obj, arq_list))
    return obj_extarqs


def get_fake_extarq_resolved_objs():
    arq_list = _get_arqs_resloved_as_dict()
    obj_extarqs = list(map(_convert_from_dict_to_obj, arq_list))
    return obj_extarqs


def get_fake_extarq_bind_objs():
    arq_list = _get_arqs_bind_as_dict()
    obj_extarqs = list(map(_convert_from_dict_to_obj, arq_list))
    return obj_extarqs


def _convert_from_dict_to_fpga_obj(arq_dict):
    obj_arq = arq.ARQ()
    for field in arq_dict.keys():
        obj_arq[field] = arq_dict[field]
    obj_extarq = fpga_ext_arq.FPGAExtARQ()
    obj_extarq.arq = obj_arq
    obj_extarq.device_profile_group = arq_dict["device_profile_group"]
    return obj_extarq


def get_fake_fpga_extarq_objs():
    arq_list = _get_arqs_as_dict()[1:]
    obj_extarqs = list(map(_convert_from_dict_to_fpga_obj, arq_list))
    return obj_extarqs


def get_fake_db_extarqs():
    db_extarqs = []
    for db_extarq in _get_arqs_as_dict():
        db_extarq.update({'device_profile_id': 0})
        db_extarqs.append(db_extarq)
    return db_extarqs


def get_fake_fpga_db_extarqs():
    return get_fake_db_extarqs()[1:]


def get_patch_list(same_device=True):
    """Returns a list of bindings for many ARQs.

    :param same_device: Flag that the returned bindings for all ARQs
        must be for the same device.
    """
    arqs = _get_arqs_as_dict()
    host_binding = {'path': '/hostname', 'op': 'add',
                    'value': arqs[0]['hostname']}
    inst_binding = {'path': '/instance_uuid', 'op': 'add',
                    'value': arqs[0]['instance_uuid']}
    device_rp_uuid = 'fb16c293-5739-4c84-8590-926f9ab16669'
    patch_list = {}
    for newarq in arqs:
        dev_uuid = device_rp_uuid if same_device else newarq['device_rp_uuid']
        dev_binding = {'path': '/device_rp_uuid', 'op': 'add',
                       'value': dev_uuid}
        patch_list[newarq['uuid']] = [host_binding, inst_binding, dev_binding]
    return patch_list, device_rp_uuid


def get_fake_xilinx_fpga_extarq_objs():
    arqs = [
        {"uuid": 'b8c19eb2-e03c-47b4-b7cf-ced6086b2d11',
         "device_profile_group_id": 0,
         "state": "Initial",
         "device_profile_name": "fake_xilinx_fpga_dp",
         "hostname": "myhost",
         "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
         "attach_handle_type": "PCI",
         # attach_handle info should vary across ARQs but ignored for testing
         "attach_handle_info": {
             "bus": "3b",
             "device": "00",
             "domain": "0000",
             "function": "0"
         },
         "device_profile_group": {
             "trait:CUSTOM_FPGA_XILINX": "required",
             "resources:FPGA": "1",
             "trait:CUSTOM_FPGA_PRODUCT_ID_5000": "required"}
         },
        {"uuid": '012955c7-90f9-45a9-bb7d-7c2907d8997f',
         "device_profile_group_id": 1,
         "state": "Initial",
         "device_profile_name": "fake_xilinx_fpga_dp",
         "hostname": "myhost",
         "instance_uuid": "5922a70f-1e06-4cfd-88dd-a332120d7144",
         "attach_handle_type": "PCI",
         # attach_handle info should vary across ARQs but ignored for testing
         "attach_handle_info": {
             "bus": "3b",
             "device": "00",
             "domain": "0000",
             "function": "1"
         },
         "device_profile_group": {
             "trait:CUSTOM_FPGA_XILINX": "required",
             "resources:FPGA": "1",
             "trait:CUSTOM_FPGA_PRODUCT_ID_5000": "required"}
         },
    ]
    return list(map(_convert_from_dict_to_obj, arqs))



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock
import webtest

from oslo_serialization import jsonutils

from cyborg.api.controllers import base
from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_device_profile


class TestDeviceProfileController(v2_test.APITestV2):
    DP_URL = '/device_profiles'

    def setUp(self):
        super(TestDeviceProfileController, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.fake_dp_objs = fake_device_profile.get_obj_devprofs()
        self.fake_dps = fake_device_profile.get_api_devprofs()

    def _validate_links(self, links, dp_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], dp_uuid)
        self.assertTrue(has_self_link)

    def _validate_dp(self, in_dp, out_dp):
        self.assertEqual(in_dp['name'], out_dp['name'])
        self.assertEqual(in_dp['uuid'], out_dp['uuid'])
        self.assertEqual(in_dp['groups'], out_dp['groups'])
        self.assertEqual(in_dp['description'], out_dp['description'])

        # Check that the link is properly set up
        self._validate_links(out_dp['links'], in_dp['uuid'])

    @mock.patch('cyborg.objects.DeviceProfile.get_by_uuid')
    def test_get_one_by_uuid(self, mock_dp_uuid):
        dp = self.fake_dp_objs[0]
        mock_dp_uuid.return_value = dp
        url = self.DP_URL + '/%s'
        data = self.get_json(url % dp['uuid'], headers=self.headers)
        mock_dp_uuid.assert_called_once()
        out_dp = data['device_profile']
        self._validate_dp(dp, out_dp)

    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    def test_get_one_by_name_before_v22(self, mock_dp_name):
        dp = self.fake_dp_objs[0]
        mock_dp_name.return_value = dp
        url = self.DP_URL + '/%s'
        headers = self.headers
        headers[base.Version.current_api_version] = '2.1'
        self.assertRaisesRegex(
            webtest.app.AppError,
            "Request not acceptable.*",
            self.get_json,
            url % dp['name'],
            headers=headers)

    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    def test_get_one_by_name(self, mock_dp_name):
        dp = self.fake_dp_objs[0]
        mock_dp_name.return_value = dp
        url = self.DP_URL + '/%s'
        headers = self.headers
        headers[base.Version.current_api_version] = '2.2'
        data = self.get_json(url % dp['name'],
                             headers=headers)
        mock_dp_name.assert_called_once()
        out_dp = data['device_profile']
        self._validate_dp(dp, out_dp)

    @mock.patch('cyborg.objects.DeviceProfile.list')
    def test_get_all(self, mock_dp):
        mock_dp.return_value = self.fake_dp_objs
        data = self.get_json(self.DP_URL, headers=self.headers)
        out_dps = data['device_profiles']

        result = isinstance(out_dps, list)
        self.assertTrue(result)
        self.assertEqual(len(out_dps), len(self.fake_dp_objs))
        for in_dp, out_dp in zip(self.fake_dp_objs, out_dps):
            self._validate_dp(in_dp, out_dp)

    @mock.patch('cyborg.objects.DeviceProfile.list')
    def test_get_all_by_name(self, mock_dp):
        mock_dp.return_value = self.fake_dp_objs
        name = 'dp_example_1'
        data = self.get_json(self.DP_URL + '?name=' + name,
                             headers=self.headers)
        out_dps = data['device_profiles']
        expected_dps = [dp for dp in self.fake_dp_objs if dp.name in [name]]

        result = isinstance(out_dps, list)
        self.assertTrue(result)
        self.assertEqual(len(out_dps), len(expected_dps))
        for in_dp, out_dp in zip(expected_dps, out_dps):
            self._validate_dp(in_dp, out_dp)

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_create')
    def test_create(self, mock_cond_dp):
        dp = [self.fake_dps[0]]
        mock_cond_dp.return_value = self.fake_dp_objs[0]
        dp[0]['created_at'] = str(dp[0]['created_at'])
        response = self.post_json(self.DP_URL, dp, headers=self.headers)
        out_dp = jsonutils.loads(response.controller_output)

        self.assertEqual(HTTPStatus.CREATED, response.status_int)
        self._validate_dp(dp[0], out_dp)

    def test_create_with_no_name(self):
        test_unsupported_dp = self.fake_dps[0]

        # delete dp name for test
        del test_unsupported_dp['name']
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            "DeviceProfile name needed.",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_unsupported_name(self):
        test_unsupported_dp = self.fake_dps[0]

        # generate special dp name for test
        test_unsupported_dp['name'] = '!'
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Device profile name must be of the form *",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_no_groups(self):
        test_unsupported_dp = self.fake_dps[0]

        # delete dp groups for test
        del test_unsupported_dp['groups']
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            "DeviceProfile needs groups field.",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_unsupported_group_key(self):
        test_unsupported_dp = self.fake_dps[0]

        # generate special dp group key for test
        del test_unsupported_dp['groups'][0]['resources:FPGA']
        test_unsupported_dp['groups'][0]['fake:FPGA'] = 'required'
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Device profile group keys must be of the form *",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_unsupported_trait_value(self):
        test_unsupported_dp = self.fake_dps[0]

        # generate special dp trait value for test
        test_unsupported_dp['groups'][0][
            'trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10'] = 'fake'
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Unsupported trait value fake *",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_unsupported_trait_name(self):
        test_unsupported_dp = self.fake_dps[0]

        # generate special trait for test
        del test_unsupported_dp['groups'][0][
            'trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10']
        test_unsupported_dp['groups'][0]['trait:FAKE_TRAIT'] = 'required'
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Unsupported trait name format FAKE_TRAIT.*",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_create')
    def test_create_with_extra_space_in_trait(self, mock_cond_dp):
        test_unsupported_dp = self.fake_dps[0]

        # generate a requested dp which has extra space in trait
        del test_unsupported_dp['groups'][0][
            'trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10']
        test_unsupported_dp['groups'][0][
            'trait:  CUSTOM_FPGA_INTEL_PAC_ARRIA10'] = 'required'

        mock_cond_dp.return_value = self.fake_dp_objs[0]
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])

        response = self.post_json(
            self.DP_URL, [test_unsupported_dp], headers=self.headers)
        out_dp = jsonutils.loads(response.controller_output)

        # check that the extra space in trait:
        # {'trait:  CUSTOM_FPGA_INTEL_PAC_ARRIA10': 'required'} is
        # successful stripped by the _validate_post_request function, and
        # the created device_profile has no extra space in trait
        # {'trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10': 'required}
        self.assertTrue(out_dp['groups'] == self.fake_dp_objs[0]['groups'])

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_create')
    def test_create_with_extra_space_in_rc(self, mock_cond_dp):
        test_unsupported_dp = self.fake_dps[0]

        # generate a requested dp which has extra space in rc
        del test_unsupported_dp['groups'][0]['resources:FPGA']
        test_unsupported_dp['groups'][0]['resources: FPGA '] = '1'

        mock_cond_dp.return_value = self.fake_dp_objs[0]
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])

        response = self.post_json(
            self.DP_URL, [test_unsupported_dp], headers=self.headers)
        out_dp = jsonutils.loads(response.controller_output)

        # check that the extra space in rc:{'resources: FPGA ': '1'} is
        # successful stripped by the _validate_post_request function, and
        # the created device_profile has no extra space in
        # rc:{'resources:FPGA': '1'}
        self.assertTrue(out_dp['groups'] == self.fake_dp_objs[0]['groups'])

    def test_create_with_unsupported_rc(self):
        test_unsupported_dp = self.fake_dps[0]
        # generate a special rc for test
        del test_unsupported_dp['groups'][0]['resources:FPGA']
        test_unsupported_dp['groups'][0]["resources:FAKE_RC"] = '1'

        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Unsupported resource class FAKE_RC.*",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    def test_create_with_invalid_resource_value(self):
        test_unsupported_dp = self.fake_dps[0]
        del test_unsupported_dp['groups'][0]['resources:FPGA']
        test_unsupported_dp['groups'][0]["resources:CUSTOM_FAKE_RC"] = 'fake'
        test_unsupported_dp['created_at'] = str(
            test_unsupported_dp['created_at'])
        self.assertRaisesRegex(
            webtest.app.AppError,
            ".*Resources number fake is invalid.*",
            self.post_json,
            self.DP_URL,
            [test_unsupported_dp],
            headers=self.headers)

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_delete')
    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    @mock.patch('cyborg.objects.DeviceProfile.get_by_uuid')
    def test_delete(self, mock_dp_uuid, mock_dp_name, mock_cond_del):
        # Delete by UUID
        url = self.DP_URL + "/5d2c0797-c3cd-4f4b-b0d0-2cc5e99ef66e"
        response = self.delete(url, headers=self.headers)
        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)
        # Delete by name
        url = self.DP_URL + "/mydp"
        response = self.delete(url, headers=self.headers)
        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Run storage database migration.
"""

import sys

from oslo_config import cfg

from cyborg.common.i18n import _
from cyborg.common import service
from cyborg.conf import CONF
from cyborg.db import migration


class DBCommand(object):

    def upgrade(self):
        migration.upgrade(CONF.command.revision)

    def revision(self):
        migration.revision(CONF.command.message, CONF.command.autogenerate)

    def stamp(self):
        migration.stamp(CONF.command.revision)

    def version(self):
        print(migration.version())

    def create_schema(self):
        migration.create_schema()


def add_command_parsers(subparsers):
    command_object = DBCommand()

    parser = subparsers.add_parser(
        'upgrade',
        help=_("Upgrade the database schema to the latest version. "
               "Optionally, use --revision to specify an alembic revision "
               "string to upgrade to."))
    parser.set_defaults(func=command_object.upgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser(
        'revision',
        help=_("Create a new alembic revision. "
               "Use --message to set the message string."))
    parser.set_defaults(func=command_object.revision)
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')

    parser = subparsers.add_parser('stamp')
    parser.set_defaults(func=command_object.stamp)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser(
        'version',
        help=_("Print the current version information and exit."))
    parser.set_defaults(func=command_object.version)

    parser = subparsers.add_parser(
        'create_schema',
        help=_("Create the database schema."))
    parser.set_defaults(func=command_object.create_schema)


def main():
    command_opt = cfg.SubCommandOpt('command',
                                    title='Command',
                                    help=_('Available commands'),
                                    handler=add_command_parsers)

    CONF.register_cli_opt(command_opt)

    service.prepare_service(sys.argv)
    CONF.command.func()



# Copyright 2023 Inspur, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

from oslo_serialization import jsonutils

from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_attribute


class TestAttributes(v2_test.APITestV2):
    ATTRIBUTE_URL = '/attributes'

    def setUp(self):
        super(TestAttributes, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.fake_attributes = fake_attribute.fake_db_attribute()
        self.fake_attribute_objs = \
            [fake_attribute.fake_attribute_obj(self.context)]

    def _validate_links(self, links, attribute_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], attribute_uuid)
        self.assertTrue(has_self_link)

    def _validate_attributes(self, in_attributes, out_attributes):
        self.assertEqual(in_attributes['uuid'], out_attributes['uuid'])
        self.assertEqual(in_attributes['key'], out_attributes['key'])
        self.assertEqual(in_attributes['value'], out_attributes['value'])
        self.assertEqual(in_attributes['deployable_id'],
                         out_attributes['deployable_id'])

        # Check that the link is properly set up
        self._validate_links(out_attributes['links'], in_attributes['uuid'])

    @mock.patch('cyborg.objects.Attribute.get')
    def test_get_one_by_uuid(self, mock_attributes_uuid):
        attribute = self.fake_attribute_objs[0]
        mock_attributes_uuid.return_value = attribute
        url = self.ATTRIBUTE_URL + '/%s'
        out_attribute = self.get_json(url % attribute['uuid'],
                                      headers=self.headers)
        mock_attributes_uuid.assert_called_once()
        self._validate_attributes(attribute, out_attribute)

    @mock.patch('cyborg.objects.Attribute.get_by_filter')
    def test_get_attribute_by_deployable_id(self, mock_deployable_id):
        attributes = self.fake_attribute_objs
        mock_deployable_id.return_value = attributes
        deployable_id = attributes[0]['deployable_id']
        url = self.ATTRIBUTE_URL + "?deployable_id=" + str(deployable_id)
        out_attributes = self.get_json(url, headers=self.headers)
        mock_deployable_id.assert_called_once()
        self.assertEqual(len(out_attributes), len(attributes))

    @mock.patch('cyborg.objects.Attribute.get_by_filter')
    def test_get_all(self, mock_attributes):
        mock_attributes.return_value = self.fake_attribute_objs
        data = self.get_json(self.ATTRIBUTE_URL, headers=self.headers)
        out_attributes = data['attributes']
        self.assertIsInstance(out_attributes, list)
        self.assertEqual(len(out_attributes), len(self.fake_attribute_objs))
        for in_attribute, out_attribute in zip(self.fake_attribute_objs,
                                               out_attributes):
            self._validate_attributes(in_attribute, out_attribute)

    @mock.patch('cyborg.objects.Attribute.get_by_filter')
    def test_get_attribute_by_key(self, mock_key):
        attributes = self.fake_attribute_objs
        mock_key.return_value = attributes
        url = self.ATTRIBUTE_URL + "?key=" + attributes[0]['key']
        out_attributes = self.get_json(url, headers=self.headers)
        mock_key.assert_called_once()
        self.assertEqual(len(out_attributes), len(attributes))

    @mock.patch('cyborg.objects.Attribute.create')
    def test_create(self, mock_cond_attribute):
        mock_cond_attribute.return_value = self.fake_attribute_objs[0]
        response = self.post_json(self.ATTRIBUTE_URL, self.fake_attributes,
                                  headers=self.headers)
        out_attribute = jsonutils.loads(response.controller_output)
        self.assertEqual(HTTPStatus.CREATED, response.status_int)
        self._validate_attributes(self.fake_attribute_objs[0], out_attribute)

    @mock.patch('cyborg.objects.Attribute.get_by_filter')
    def test_get_all_by_deployable_id_and_key(self, mock_attribute):
        attributes = self.fake_attribute_objs
        mock_attribute.return_value = attributes
        dp_url = "?deployable_id=" + str(attributes[0]['deployable_id'])
        url = self.ATTRIBUTE_URL + dp_url + "&key=" + attributes[0]['key']
        data = self.get_json(url, headers=self.headers)
        out_attributes = data['attributes']
        mock_attribute.assert_called_once()
        self.assertEqual(len(out_attributes), len(attributes))
        for in_attribute, out_attribute in zip(attributes, out_attributes):
            self._validate_attributes(in_attribute, out_attribute)

    @mock.patch('cyborg.objects.Attribute.get')
    @mock.patch('cyborg.objects.Attribute.destroy')
    def test_delete(self, mock_attribute_delete, mock_attribute):
        uuid = self.fake_attribute_objs[0]['uuid']
        # Delete by UUID
        url = self.ATTRIBUTE_URL + '/%s' % uuid
        response = self.delete(url, headers=self.headers)
        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)



# Copyright 2017 Lenovo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Generic driver modules implementation.
"""

from cyborg.accelerator.common import exception
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

# NOTE (crushil): REQUIRED_PROPERTIES needs to be filled out.
REQUIRED_PROPERTIES = {'create', 'get', 'list', 'update', 'delete'}
COMMON_PROPERTIES = REQUIRED_PROPERTIES


def _check_for_missing_params(info_dict, error_msg, param_prefix=''):
    missing_info = []
    for label, value in info_dict.items():
        if not value:
            missing_info.append(param_prefix + label)

    if missing_info:
        exc_msg = _("%(error_msg)s. Missing are: %(missing_info)s")
        raise exception.MissingParameterValue(
            exc_msg % {'error_msg': error_msg, 'missing_info': missing_info})


def _parse_driver_info(driver):
    info = driver.driver_info
    d_info = {k: info.get(k) for k in COMMON_PROPERTIES}
    error_msg = _("Cannot validate Generic Driver. Some parameters were"
                  " missing in the configuration file.")
    _check_for_missing_params(d_info, error_msg)
    return d_info


class GENERICDRIVER(object):

    def get_properties(self):
        """Return the properties of the generic driver.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return COMMON_PROPERTIES

    def attach(self, accelerator, instance):

        def install(self, accelerator):
            pass

    def detach(self, accelerator, instance):

        def uninstall(self, accelerator):
            pass

        def delete(self):
            pass

    def discover(self):
        pass

    def list(self):
        pass

    def update(self, accelerator, **kwargs):
        pass



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg FPGA driver implementation.
"""

from cyborg.accelerator.drivers.fpga import utils


VENDOR_MAPS = {"0x8086": "intel",
               "1bd4": 'inspur'}


class FPGADriver(object):
    """Base class for FPGA drivers.

       This is just a virtual FPGA drivers interface.
       Vendor should implement their specific drivers.
    """

    @classmethod
    def create(cls, vendor, *args, **kwargs):
        for sclass in cls.__subclasses__():
            vendor = VENDOR_MAPS.get(vendor, vendor)
            if vendor == sclass.VENDOR:
                return sclass(*args, **kwargs)
        raise LookupError("Not find the FPGA driver for vendor %s" % vendor)

    def __init__(self, *args, **kwargs):
        pass

    def discover(self):
        raise NotImplementedError()

    def program(self, device_path, image):
        raise NotImplementedError()

    @classmethod
    def discover_vendors(cls):
        return utils.discover_vendors()



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from oslo_log import log as logging
from oslo_serialization import jsonutils
from unittest import mock

from cyborg.api.controllers.v2 import device_profiles

from cyborg.tests.unit import fake_device_profile
from cyborg.tests.unit.policies import base


LOG = logging.getLogger(__name__)
DP_URL = '/device_profiles'


class DeviceProfilePolicyTest(base.BasePolicyTest):
    """Test device_profile APIs policies with all possible contexts.

    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(DeviceProfilePolicyTest, self).setUp()

        self.flags(enforce_scope=False, group="oslo_policy")
        self.controller = device_profiles.DeviceProfilesController()
        self.fake_dp_objs = fake_device_profile.get_obj_devprofs()
        self.fake_dps = fake_device_profile.get_api_devprofs()
        # check both legacy and new policies for create APIs
        self.create_authorized_contexts = [
            self.legacy_admin_context,  # legacy: admin
            self.system_admin_context,  # new policy: system_admin
            self.project_admin_context
        ]
        self.create_unauthorized_contexts = list(
            set(self.all_contexts) - set(self.create_authorized_contexts))

        # check both legacy and new policies for delete APIs
        self.delete_authorized_contexts = [
            # legacy: admin_or_owner
            self.legacy_admin_context,
            # NOTE(yumeng) although the legacy rule is admin_or_owner,
            # cyborg does not "actually" support owner check for
            # device profile, so we just uncomment legacy_owner_context here.
            # If later we need support owner policy, we should recheck here.
            # self.legacy_owner_context,
            self.system_admin_context,  # new policy: system_admin
            self.project_admin_context
        ]
        self.delete_unauthorized_contexts = list(
            set(self.all_contexts) - set(self.delete_authorized_contexts))

    def _validate_links(self, links, dp_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], dp_uuid)
        self.assertTrue(has_self_link)

    def _validate_dp(self, in_dp, out_dp):
        self.assertEqual(in_dp['name'], out_dp['name'])
        self.assertEqual(in_dp['uuid'], out_dp['uuid'])
        self.assertEqual(in_dp['groups'], out_dp['groups'])

        # Check that the link is properly set up
        self._validate_links(out_dp['links'], in_dp['uuid'])

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_create')
    def test_create_device_profile_success(self, mock_cond_dp):
        mock_cond_dp.return_value = self.fake_dp_objs[0]
        dp = [self.fake_dps[0]]
        dp[0]['created_at'] = str(dp[0]['created_at'])

        for context in self.create_authorized_contexts:
            headers = self.gen_headers(context)
            response = self.post_json(DP_URL, dp, headers=headers)
            out_dp = jsonutils.loads(response.controller_output)

            self.assertEqual(HTTPStatus.CREATED, response.status_int)
            self._validate_dp(dp[0], out_dp)

    def test_create_device_profile_forbidden(self):
        dp = [self.fake_dps[0]]
        dp[0]['created_at'] = str(dp[0]['created_at'])
        for context in self.create_unauthorized_contexts:
            headers = self.gen_headers(context)
            try:
                self.post_json(DP_URL, dp, headers=headers)
            except Exception as e:
                exc = e
            self.assertIn("Bad response: 403 Forbidden", exc.args[0])

    @mock.patch('cyborg.conductor.rpcapi.ConductorAPI.device_profile_delete')
    @mock.patch('cyborg.objects.DeviceProfile.get_by_name')
    @mock.patch('cyborg.objects.DeviceProfile.get_by_uuid')
    def test_delete_device_profile_success(self, mock_dp_uuid,
                                           mock_dp_name, mock_cond_del):
        for context in self.delete_authorized_contexts:
            headers = self.gen_headers(context)
            # Delete by UUID
            url = DP_URL + "/5d2c0797-c3cd-4f4b-b0d0-2cc5e99ef66e"
            response = self.delete(url, headers=headers)
            self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)
            # Delete by name
            url = DP_URL + "/mydp"
            response = self.delete(url, headers=headers)
            self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)

    def test_delete_device_profile_forbidden(self):
        dp = self.fake_dp_objs[0]
        url = DP_URL + '/%s'
        exc = None
        for context in self.delete_unauthorized_contexts:
            headers = self.gen_headers(context)
            try:
                self.delete(url % dp['uuid'], headers=headers)
            except Exception as e:
                exc = e
            self.assertIn("Bad response: 403 Forbidden", exc.args[0])


class DeviceProfileScopeTypePolicyTest(DeviceProfilePolicyTest):
    """Test device_profile APIs policies with system scope enabled.
    This class set the cyborg.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(DeviceProfileScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")
        # check that admin is able to do create and delete operations.
        self.create_authorized_contexts = [
            self.legacy_admin_context,
            self.project_admin_context]
        self.delete_authorized_contexts = self.create_authorized_contexts
        # Check that system or non-admin is not able to perform the system
        # level actions on device_profiles.
        self.create_unauthorized_contexts = [
            self.system_admin_context, self.system_member_context,
            self.system_reader_context, self.system_foo_context,
            self.project_member_context,
            self.other_project_member_context,
            self.project_foo_context, self.project_reader_context
        ]
        self.delete_unauthorized_contexts = self.create_unauthorized_contexts



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""SQLAlchemy models for accelerator service."""

from oslo_db import options as db_options
from oslo_db.sqlalchemy import models
from oslo_utils import timeutils
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import String
from sqlalchemy import Text
import urllib.parse as urlparse

from cyborg.common import constants
from cyborg.common import paths
from cyborg.conf import CONF


_DEFAULT_SQL_CONNECTION = 'sqlite:///' + paths.state_path_def('cyborg.sqlite')
db_options.set_defaults(CONF, connection=_DEFAULT_SQL_CONNECTION)


def table_args():
    engine_name = urlparse.urlparse(CONF.database.connection).scheme
    if engine_name == 'mysql':
        return {'mysql_engine': CONF.database.mysql_engine,
                'mysql_charset': "utf8"}
    return None


class CyborgBase(models.TimestampMixin, models.ModelBase):
    metadata = None

    def as_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d

    @staticmethod
    def delete_values():
        return {'deleted': True,
                'deleted_at': timeutils.utcnow()}

    def delete(self, session):
        """Delete this object."""
        updated_values = self.delete_values()
        self.update(updated_values)
        self.save(session=session)
        return updated_values


Base = declarative_base(cls=CyborgBase)


class Device(Base):
    """Represents the devices."""

    __tablename__ = 'devices'

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    type = Column(Enum('GPU', 'FPGA', 'AICHIP', 'QAT', 'NIC', 'SSD',
                       name='device_type'), nullable=False)
    vendor = Column(String(255), nullable=False)
    model = Column(String(255), nullable=False)
    std_board_info = Column(Text, nullable=True)
    vendor_board_info = Column(Text, nullable=True)
    hostname = Column(String(255), nullable=False)
    status = Column(Enum("enabled", "maintaining", name='device_status'),
                    default='enabled', nullable=False)


class Deployable(Base):
    """Represents the deployables."""

    __tablename__ = 'deployables'
    __table_args__ = (
        Index('deployables_parent_id_idx', 'parent_id'),
        Index('deployables_root_id_idx', 'root_id'),
        Index('deployables_device_id_idx', 'device_id'),
        table_args()
    )

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    parent_id = Column(Integer, ForeignKey('deployables.id'), nullable=True)
    root_id = Column(Integer, ForeignKey('deployables.id'), nullable=True)
    name = Column(String(255), nullable=False)
    num_accelerators = Column(Integer, nullable=False)
    device_id = Column(Integer, ForeignKey('devices.id', ondelete="RESTRICT"),
                       nullable=False)
    # The resource provider UUID is nullable for 2 reasons:
    # A. on creation, till Placement is populated, this will be null.
    # B. Sub-deployables (such as in networked FPGA cards) will have
    #    this as null.
    rp_uuid = Column(String(36), nullable=True)
    # This is nullable because the parent deployable in a networked
    # FPGA card will have no driver. Only subdeployables have one.
    driver_name = Column(String(100), nullable=True)
    bitstream_id = Column(String(36), nullable=True)

    # TODO() Add programming_in_progress field


class Attribute(Base):
    __tablename__ = 'attributes'

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    deployable_id = Column(Integer,
                           ForeignKey('deployables.id', ondelete="RESTRICT"),
                           nullable=False, index=True)
    key = Column(Text, nullable=False)
    value = Column(Text, nullable=False)


class ControlpathID(Base):
    """Identifier for the Device when driver reporting to agent, IDs is
    needed especially when multiple PFs exist in one Devices.
    """

    __tablename__ = 'controlpath_ids'

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    device_id = Column(Integer,
                       ForeignKey('devices.id', ondelete="RESTRICT"),
                       nullable=False, index=True)
    cpid_type = Column(Enum('PCI', name='cpid_type'), nullable=False)
    cpid_info = Column(String(255), nullable=False)


class AttachHandle(Base):
    """Represents device's VFs and PFs which can be attached to a VM."""

    __tablename__ = 'attach_handles'
    __table_args__ = (
        Index('attach_handles_cpid_id_idx', 'cpid_id'),
        Index('attach_handles_deployable_id_idx', 'deployable_id'),
        table_args()
    )

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    deployable_id = Column(Integer,
                           ForeignKey('deployables.id', ondelete="RESTRICT"),
                           nullable=False)
    cpid_id = Column(Integer,
                     ForeignKey('controlpath_ids.id', ondelete="RESTRICT"),
                     nullable=False)
    in_use = Column(Boolean, default=False)
    attach_type = Column(Enum(constants.AH_TYPE_PCI,
                              constants.AH_TYPE_MDEV,
                              constants.AH_TYPE_TEST_PCI,
                         name='attach_type'),
                         nullable=False)
    attach_info = Column(String(255), nullable=False)


class DeviceProfile(Base):
    """Represents users' specific requirements."""

    __tablename__ = 'device_profiles'
    __table_args__ = (
        schema.UniqueConstraint('uuid', 'name',
                                name='uniq_device_profiles0uuid0name'),
        table_args()
    )

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=False, unique=True)
    profile_json = Column(Text, nullable=False)
    description = Column(Text, nullable=True)


class ExtArq(Base):
    """ExtArq is the abbreviation of ExtendedAcceleratorRequest, it represents
    extended nova requests for attach related operations.
    """

    __tablename__ = 'extended_accelerator_requests'
    __table_args__ = (
        Index('extArqs_project_id_idx', 'project_id'),
        Index('extArqs_device_profile_id_idx', 'device_profile_id'),
        Index('extArqs_device_rp_uuid_idx', 'device_rp_uuid'),
        Index('extArqs_instance_uuid_idx', 'instance_uuid'),
        Index('extArqs_attach_handle_id_idx', 'attach_handle_id'),
        Index('extArqs_deployable_id_idx', 'deployable_id'),
        table_args()
    )

    id = Column(Integer, primary_key=True)
    uuid = Column(String(36), nullable=False, unique=True)
    project_id = Column(String(255), nullable=True)
    state = Column(Enum(constants.ARQ_INITIAL,
                        constants.ARQ_BIND_STARTED,
                        constants.ARQ_BOUND,
                        constants.ARQ_BIND_FAILED,
                        constants.ARQ_UNBOUND,
                        constants.ARQ_DELETING),
                   nullable=False)
    device_profile_id = Column(Integer, ForeignKey('device_profiles.id',
                                                   ondelete="RESTRICT"),
                               nullable=False)
    device_profile_group_id = Column(Integer, nullable=False, default=0)
    hostname = Column(String(255), nullable=True)
    device_rp_uuid = Column(String(36), nullable=True)
    instance_uuid = Column(String(36), nullable=True)
    attach_handle_id = Column(Integer, ForeignKey('attach_handles.id',
                                                  ondelete="RESTRICT"),
                              nullable=True)
    # Cyborg Private Fields
    substate = Column(Enum('Initial', name='substate'), nullable=False,
                      default='Initial')
    deployable_id = Column(Integer,
                           ForeignKey('deployables.id', ondelete="RESTRICT"),
                           nullable=True)


class QuotaUsage(Base):
    """Represents the current usage for a given resource."""

    __tablename__ = 'quota_usages'
    __table_args__ = (
        Index('ix_quota_usages_project_id', 'project_id'),
        Index('ix_quota_usages_user_id', 'user_id'),
    )
    id = Column(Integer, primary_key=True)

    project_id = Column(String(255))
    user_id = Column(String(255))
    resource = Column(String(255), nullable=False)

    in_use = Column(Integer, nullable=False)
    reserved = Column(Integer, nullable=False)

    @property
    def total(self):
        return self.in_use + self.reserved

    until_refresh = Column(Integer)


class Reservation(Base):
    """Represents a resource reservation for quotas."""

    __tablename__ = 'reservations'
    __table_args__ = (
        Index('ix_reservations_project_id', 'project_id'),
        Index('reservations_uuid_idx', 'uuid'),
        Index('ix_reservations_user_id', 'user_id'),
    )
    id = Column(Integer, primary_key=True, nullable=False)
    uuid = Column(String(36), nullable=False)

    usage_id = Column(Integer, ForeignKey('quota_usages.id'), nullable=False)

    project_id = Column(String(255))
    user_id = Column(String(255))
    resource = Column(String(255))

    delta = Column(Integer, nullable=False)
    expire = Column(DateTime)

    usage = orm.relationship(
        "QuotaUsage",
        foreign_keys=usage_id,
        primaryjoin=usage_id == QuotaUsage.id)



# Copyright 2018 Huawei Technologies Co.,LTD.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg import objects
from cyborg.objects import device
from cyborg.objects import fields


def get_fake_devices_as_dict():
    device1 = {
        "id": 1,
        "vendor": "0xABCD",
        "uuid": "1c6c9033-560d-4a7a-bb8e-94455d1e7825",
        "hostname": "test-node-1",
        "vendor_board_info": "fake_vendor_info",
        "model": "miss model info",
        "type": "FPGA",
        "std_board_info": "{'class': 'Fake class', 'device_id': '0xabcd'}"
        }
    device2 = {
        "id": 2,
        "vendor": "0xDCBA",
        "uuid": "1c6c9033-560d-4a7a-bb8e-94455d1e7826",
        "hostname": "test-node-2",
        "vendor_board_info": "fake_vendor_info",
        "model": "miss model info",
        "type": "GPU",
        "std_board_info": "{'class': 'Fake class', 'device_id': '0xdcba'}"
        }
    return [device1, device2]


def _convert_from_dict_to_obj(device_dict):
    obj_device = device.Device()
    for field in device_dict.keys():
        obj_device[field] = device_dict[field]
    return obj_device


def _convert_to_db_device(device_dict):
    for name, field in objects.Device.fields.items():
        if name in device_dict:
            continue
        if field.nullable:
            device_dict[name] = None
        elif field.default != fields.UnspecifiedDefault:
            device_dict[name] = field.default
        else:
            raise Exception('fake_db_device needs help with %s' % name)
    return device_dict


def get_db_devices():
    devices_list = get_fake_devices_as_dict()
    db_devices = list(map(_convert_to_db_device, devices_list))
    return db_devices


def get_fake_devices_objs():
    devices_list = get_fake_devices_as_dict()
    obj_devices = list(map(_convert_from_dict_to_obj, devices_list))
    return obj_devices



#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for Magnum specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range M3xx. Find the current test with
   the highest allocated number and then pick the next value.
   If nova has an N3xx code for that test, use the same number.
 - Keep the test method code in the source file ordered based
   on the M3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to magnum/tests/unit/test_hacking.py

"""
UNDERSCORE_IMPORT_FILES = []

mutable_default_args = re.compile(r"^\s*def .+\((.+=\{\}|.+=\[\])")
assert_equal_in_end_with_true_or_false_re = re.compile(
    r"assertEqual\((\w|[][.'\"])+ in (\w|[][.'\", ])+, (True|False)\)")
assert_equal_in_start_with_true_or_false_re = re.compile(
    r"assertEqual\((True|False), (\w|[][.'\"])+ in (\w|[][.'\", ])+\)")
assert_equal_with_is_not_none_re = re.compile(
    r"assertEqual\(.*?\s+is+\s+not+\s+None\)$")
assert_true_isinstance_re = re.compile(
    r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
    r"(\w|\.|\'|\"|\[|\])+\)\)")
dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")
assert_xrange_re = re.compile(
    r"\s*xrange\s*\(")
log_translation = re.compile(
    r"(.)*LOG\.(audit|error|critical)\(\s*('|\")")
log_translation_info = re.compile(
    r"(.)*LOG\.(info)\(\s*(_\(|'|\")")
log_translation_exception = re.compile(
    r"(.)*LOG\.(exception)\(\s*(_\(|'|\")")
log_translation_LW = re.compile(
    r"(.)*LOG\.(warning|warn)\(\s*(_\(|'|\")")
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")
underscore_import_check = re.compile(r"(.)*import _(.)*")
translated_log = re.compile(
    r"(.)*LOG\.(audit|error|info|critical|exception)"
    r"\(\s*_\(\s*('|\")")
string_translation = re.compile(r"[^_]*_\(\s*('|\")")


@core.flake8ext
def no_mutable_default_args(logical_line):
    msg = "M322: Method's default argument shouldn't be mutable!"
    if mutable_default_args.match(logical_line):
        yield (0, msg)


@core.flake8ext
def assert_equal_not_none(logical_line):
    """Check for assertEqual(A is not None) sentences M302"""
    msg = "M302: assertEqual(A is not None) sentences not allowed."
    res = assert_equal_with_is_not_none_re.search(logical_line)
    if res:
        yield (0, msg)


@core.flake8ext
def assert_true_isinstance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    M316
    """
    if assert_true_isinstance_re.match(logical_line):
        yield (0, "M316: assertTrue(isinstance(a, b)) sentences not allowed")


@core.flake8ext
def assert_equal_in(logical_line):
    """Check for assertEqual(True|False, A in B),
     assertEqual(A in B, True|False)

    M338
    """
    res = (assert_equal_in_start_with_true_or_false_re.search(logical_line) or
           assert_equal_in_end_with_true_or_false_re.search(logical_line))
    if res:
        yield (0, "M338: Use assertIn/NotIn(A, B) rather than "
                  "assertEqual(A in B, True/False) when checking collection "
                  "contents.")


@core.flake8ext
def no_xrange(logical_line):
    """Disallow 'xrange()'

    M339
    """
    if assert_xrange_re.match(logical_line):
        yield (0, "M339: Do not use xrange().")


@core.flake8ext
def use_timeutils_utcnow(logical_line, filename):
    # tools are OK to use the standard datetime module
    if "/tools/" in filename:
        return

    msg = "M310: timeutils.utcnow() must be used instead of datetime.%s()"
    datetime_funcs = ['now', 'utcnow']
    for f in datetime_funcs:
        pos = logical_line.find('datetime.%s' % f)
        if pos != -1:
            yield (pos, msg % f)


@core.flake8ext
def dict_constructor_with_list_copy(logical_line):
    msg = ("M336: Must use a dict comprehension instead of a dict constructor"
           " with a sequence of key-value pairs."
           )
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


@core.flake8ext
def no_log_warn(logical_line):
    """Disallow 'LOG.warn('

    Deprecated LOG.warn(), instead use LOG.warning
    https://bugs.launchpad.net/magnum/+bug/1508442

    M352
    """

    msg = ("M352: LOG.warn is deprecated, please use LOG.warning!")
    if "LOG.warn(" in logical_line:
        yield (0, msg)


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate logs are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.
    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    if filename in UNDERSCORE_IMPORT_FILES:
        pass
    elif (underscore_import_check.match(logical_line) or
          custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif (translated_log.match(logical_line) or
          string_translation.match(logical_line)):
        yield (0, "M340: Found use of _() without explicit import of _ !")


@core.flake8ext
def import_stock_mock(logical_line):
    """Use python's mock, not the mock library.

    Since we `dropped support for python 2`__, we no longer need to use the
    mock library, which existed to backport py3 functionality into py2.
    Which must be done by saying::

        from unittest import mock

    ...because if you say::

        import mock

    ...you definitely will not be getting the standard library mock. That will
    always import the third party mock library. This check can be removed in
    the future (and we can start saying ``import mock`` again) if we manage to
    purge these transitive dependencies.

    .. __: https://review.opendev.org/#/c/688593/

    N366
    """
    if logical_line == 'import mock':
        yield (0, "N366: You must explicitly import python's mock: "
                  "``from unittest import mock``")



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 ZTE Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as object_base

from cyborg.common import constants
from cyborg.db import api as dbapi
from cyborg.objects import base
from cyborg.objects import fields as object_fields


LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class Device(base.CyborgObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Add AICHIP, GENERIC type
    # Version 1.2: Add status field
    VERSION = '1.2'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(nullable=False),
        'uuid': object_fields.UUIDField(nullable=False),
        'type': object_fields.EnumField(valid_values=constants.DEVICE_TYPE,
                                        nullable=False),
        'vendor': object_fields.StringField(nullable=False),
        'model': object_fields.StringField(nullable=False),
        'std_board_info': object_fields.StringField(nullable=True),
        'vendor_board_info': object_fields.StringField(nullable=True),
        'hostname': object_fields.StringField(nullable=False),
        'status': object_fields.EnumField(valid_values=constants.DEVICE_STATUS,
                                          nullable=False, default="enabled"),
    }

    def create(self, context):
        """Create a device record in the DB."""
        values = self.obj_get_changes()
        db_device = self.dbapi.device_create(context, values)
        self._from_db_object(self, db_device)

    @classmethod
    def get(cls, context, uuid):
        """Find a DB Device and return an Obj Device."""
        db_device = cls.dbapi.device_get(context, uuid)
        obj_device = cls._from_db_object(cls(context), db_device)
        return obj_device

    @classmethod
    def list(cls, context, filters=None):
        """Return a list of Device objects."""
        if filters:
            sort_dir = filters.pop('sort_dir', 'desc')
            sort_key = filters.pop('sort_key', 'created_at')
            limit = filters.pop('limit', None)
            marker = filters.pop('marker_obj', None)
            db_devices = cls.dbapi.device_list_by_filters(
                context, filters, sort_dir=sort_dir, sort_key=sort_key,
                limit=limit, marker=marker)
        else:
            db_devices = cls.dbapi.device_list(context)
        return cls._from_db_object_list(db_devices, context)

    def save(self, context):
        """Update a Device record in the DB."""
        updates = self.obj_get_changes()
        db_device = self.dbapi.device_update(context, self.uuid, updates)
        self._from_db_object(self, db_device)

    def destroy(self, context):
        """Delete the Device from the DB."""
        self.dbapi.device_delete(context, self.uuid)
        self.obj_reset_changes()

    @classmethod
    def get_list_by_hostname(cls, context, hostname):
        """get device object list from the hostname. return [] if not
        exist.
        """
        dev_filter = {'hostname': hostname}
        device_obj_list = Device.list(context, dev_filter)
        return device_obj_list

    @classmethod
    def get_by_device_id(cls, context, device_id):
        """get device object by the device ID."""
        db_device = cls.dbapi.device_get_by_id(context, device_id)
        obj_device = cls._from_db_object(cls(context), db_device)
        return obj_device



# -*- coding: utf-8 -*-

#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Client side of the Agent RPC API."""

from oslo_config import cfg
import oslo_messaging as messaging

from cyborg.common import constants
from cyborg.common import rpc
from cyborg.objects import base as objects_base

from oslo_log import log

LOG = log.getLogger(__name__)


CONF = cfg.CONF


class AgentAPI(object):
    """Client side of the Agent RPC API.

    API version history:

    |    1.0 - Initial version.

    """

    RPC_API_VERSION = '1.0'

    def __init__(self, topic=None):
        super(AgentAPI, self).__init__()
        self.topic = topic or constants.AGENT_TOPIC
        target = messaging.Target(topic=self.topic,
                                  version='1.0')
        serializer = objects_base.CyborgObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=self.RPC_API_VERSION,
                                     serializer=serializer)

    def fpga_program(self, context, hostname, controlpath_id,
                     bitstream_uuid, driver_name):
        LOG.info('Agent fpga_program: hostname: (%s) ' +
                 'bitstream_id: (%s)', hostname, bitstream_uuid)
        version = '1.0'
        cctxt = self.client.prepare(server=hostname, version=version)
        return cctxt.call(context, 'fpga_program',
                          controlpath_id=controlpath_id,
                          bitstream_uuid=bitstream_uuid,
                          driver_name=driver_name)

    def create_vgpu_mdev(self, context, hostname, pci_addr,
                         asked_type, ah_uuid):
        LOG.debug('Agent create_vgpu_mdev: hostname: (%s) , pci_address: (%s)'
                  'gpu_id: (%s)', hostname, pci_addr, ah_uuid)
        version = '1.0'
        cctxt = self.client.prepare(server=hostname, version=version)
        return cctxt.call(context, 'create_vgpu_mdev',
                          pci_addr=pci_addr,
                          asked_type=asked_type,
                          ah_uuid=ah_uuid)

    def remove_vgpu_mdev(self, context, hostname, pci_addr,
                         asked_type, ah_uuid):
        LOG.debug('Agent remove_vgpu_mdev: hostname: (%s) '
                  'gpu_id: (%s)', hostname, ah_uuid)
        version = '1.0'
        cctxt = self.client.prepare(server=hostname, version=version)
        return cctxt.call(context, 'remove_vgpu_mdev',
                          pci_addr=pci_addr,
                          asked_type=asked_type,
                          ah_uuid=ah_uuid)



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import fixtures
from unittest import mock

from cyborg.common import exception
from cyborg.common import nova_client
from cyborg.tests import base


class NovaAPITest(base.TestCase):
    wsgi_api_version = '2.82'

    def setUp(self):
        super(NovaAPITest, self).setUp()
        self.instance_uuid = '00000000-0000-0000-0000-000000000001'
        template = {'name': 'accelerator-request-bound',
                    'server_uuid': self.instance_uuid,
                    'code': 200,
                    'status': 'completed'}
        tags = ['00000000-0000-0000-0000-000000000002',
                '00000000-0000-0000-0000-000000000003']
        self.events = [dict(template, tag=tag) for tag in tags]

        self.mock_sdk = self.useFixture(fixtures.MockPatch(
            'cyborg.common.utils.get_sdk_adapter')).mock.return_value
        self.mock_log_info = self.useFixture(fixtures.MockPatch(
            'cyborg.common.nova_client.LOG.info')).mock

    def test_send_events(self):
        self.mock_sdk.post.return_value = mock.Mock(status_code=200)

        nova = nova_client.NovaAPI()
        nova._send_events(self.events)

        msg = 'Sucessfully sent events to Nova, events: %(events)s'
        self.mock_log_info.assert_called_once_with(
            msg, {'events': self.events})

    def test_send_events_422(self):
        # If Nova returns HTTP 207 with event code 422 for all events,
        # ignore it.
        resp_events = copy.deepcopy(self.events)
        for ev in resp_events:
            ev.update({'status': 'failed', 'code': 422})
        nova_resp = {'events': resp_events}
        mock_ret = mock.Mock(status_code=207)
        mock_ret.json.return_value = nova_resp
        self.mock_sdk.post.return_value = mock_ret

        nova = nova_client.NovaAPI()
        nova._send_events(self.events)

        msg = ('Ignoring Nova notification error that the instance %s is not '
               'yet associated with a host.')
        self.mock_log_info.assert_called_once_with(msg, self.instance_uuid)

    def test_send_events_with_event_code_422_exception(self):
        # If Nova returns HTTP 207 with event code 422 for some events,
        # but not all, raise an exception. This is not expected to
        # happen with current code.
        resp_events = copy.deepcopy(self.events)
        resp_events[0].update({'status': 'failed', 'code': 422})
        nova_resp = {'events': resp_events}
        mock_ret = mock.Mock(status_code=207)
        mock_ret.json.return_value = nova_resp
        self.mock_sdk.post.return_value = mock_ret

        nova = nova_client.NovaAPI()
        self.assertRaises(exception.InvalidAPIResponse,
                          nova._send_events, self.events)

    def test_send_events_with_event_code_400_exception(self):
        # If Nova returns HTTP 207 with event code 400 for some events,
        # raise an exception.
        resp_events = copy.deepcopy(self.events)
        resp_events[0].update({'status': 'failed', 'code': 400})
        nova_resp = {'events': resp_events}
        mock_ret = mock.Mock(status_code=207)
        mock_ret.json.return_value = nova_resp
        self.mock_sdk.post.return_value = mock_ret

        nova = nova_client.NovaAPI()
        self.assertRaises(exception.InvalidAPIResponse,
                          nova._send_events, self.events)

    def test_send_events_with_all_event_code_400_exception(self):
        # If Nova returns HTTP 207 with event code 400 for all events,
        # raise an exception.
        resp_events = copy.deepcopy(self.events)
        for ev in resp_events:
            ev.update({'status': 'failed', 'code': 400})
        nova_resp = {'events': resp_events}
        mock_ret = mock.Mock(status_code=207)
        mock_ret.json.return_value = nova_resp
        self.mock_sdk.post.return_value = mock_ret

        nova = nova_client.NovaAPI()
        self.assertRaises(exception.InvalidAPIResponse,
                          nova._send_events, self.events)

    def test_send_events_failure(self):
        # Nova is expected to return 200/207 but this is future-proofing.
        mock_ret = mock.Mock(status_code=400)
        mock_ret.json.return_value = {}  # Dummy response
        self.mock_sdk.post.return_value = mock_ret

        nova = nova_client.NovaAPI()
        self.assertRaises(exception.InvalidAPIResponse,
                          nova._send_events, self.events)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cyborg.common import exception
from cyborg.tests import base


class TestException(base.TestCase):
    def _get_all_exception_classes(self):
        classes = []

        base_class = exception.CyborgException

        for name in dir(exception):
            item = getattr(exception, name)
            if item == base_class:
                continue
            elif type(item) != type:
                continue
            elif not issubclass(item, base_class):
                continue
            classes.append(item)

        return classes

    def test_all_exceptions(self):
        bad_classes = []
        for cls in self._get_all_exception_classes():
            parent = cls.mro()[1]
            if parent._msg_fmt == cls._msg_fmt:
                bad_classes.append(cls.__name__)

        self.assertEqual([], bad_classes,
                         'Exception classes %s do not '
                         'set _msg_fmt' % ', '.join(bad_classes))



# Copyright 2020 ZTE Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

# All legacy policy and new policy mapping for all V2 APIs can be found
# here:https://wiki.openstack.org/wiki/Cyborg/Policy

DEPRECATED_REASON = """
Cyborg API policies are introducing new default roles with scope_type
capabilities. We will start to deprecate old policies from WALLABY release,
and are going to ignore all the old policies silently from X release.
Be sure to take these new defaults into consideration if you are
relying on overrides in your deployment for the policy API.
"""

deprecated_default = 'rule:admin_or_owner'
deprecated_is_admin = 'rule:is_admin'

ADMIN_OR_OWNER = 'rule:admin_or_owner'

DEPRECATED_ADMIN_OR_OWNER = policy.DeprecatedRule(
    name=ADMIN_OR_OWNER,
    check_str='is_admin:True or project_id:%(project_id)s',
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

deprecated_default_policies = [
    # is_public_api is set in the environment from AuthTokenMiddleware
    policy.RuleDefault(
        name='public_api',
        check_str='is_public_api:True',
        description='legacy rule of Internal flag for public API routes',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    # The policy check "@" will always accept an access. The empty list
    # (``[]``) or the empty string (``""``) is equivalent to the "@"
    policy.RuleDefault(
        name='allow',
        check_str='@',
        description='legacy rule: any access will be passed',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    # the policy check "!" will always reject an access.
    policy.RuleDefault(
        name='deny',
        check_str='!',
        description='legacy rule: all access will be forbidden',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    policy.RuleDefault(
        name='default',
        check_str='rule:admin_or_owner',
        description='Legacy rule for default rule',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    policy.RuleDefault(
        name='is_admin',
        check_str='rule:admin_api',
        description='Full read/write API access',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    policy.RuleDefault(
        name='admin_or_owner',
        check_str='is_admin:True or project_id:%(project_id)s',
        description='Admin or owner API access',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
    policy.RuleDefault(
        name='admin_or_user',
        check_str='is_admin:True or user_id:%(user_id)s',
        description='Admin or user API access',
        deprecated_for_removal=True,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.WALLABY),
]

ADMIN = 'rule:admin_api'
PROJECT_ADMIN = 'rule:project_admin_api'
PROJECT_MEMBER = 'rule:project_member_api'
PROJECT_READER = 'rule:project_reader_api'
PROJECT_MEMBER_OR_ADMIN = 'rule:project_member_or_admin'
PROJECT_READER_OR_ADMIN = 'rule:project_reader_or_admin'

# NOTE(yumeng): Keystone already support implied roles means assignment
# of one role implies the assignment of another. New defaults roles
# `reader`, `member` also has been added in bootstrap. If the bootstrap
# process is re-run, and a `reader`, `member`, or `admin` role already
# exists, a role implication chain will be created: `admin` implies
# `member` implies `reader`.
# For example: If we give access to 'reader' it means the 'admin' and
# 'member' also get access.

# NOTE(yumeng) the rules listed include both old rules and new rules.
# legacy rules list: 'public_api','allow','deny','admin_api','is_admin',
# 'admin_or_owner','admin_or_user'.
# new rules list: project_admin_api, project_member_api, project_reader_api,
# project_member_or_admin, project_reader_or_admin .
default_policies = [
    policy.RuleDefault(
        name='admin_api',
        check_str='role:admin or role:administrator',
        description='Legacy rule for cloud admin access'),
    policy.RuleDefault(
        name="project_admin_api",
        check_str="role:admin and project_id:%(project_id)s",
        description="Default rule for Project level admin APIs."),
    policy.RuleDefault(
        name="project_member_api",
        check_str="role:member and project_id:%(project_id)s",
        description="Default rule for Project level non admin APIs."),
    policy.RuleDefault(
        name="project_reader_api",
        check_str="role:reader and project_id:%(project_id)s",
        description="Default rule for Project level read only APIs."),
    policy.RuleDefault(
        "project_member_or_admin",
        "rule:project_member_api or rule:admin_api",
        "Default rule for Project Member or admin APIs.",
        deprecated_rule=DEPRECATED_ADMIN_OR_OWNER),
    policy.RuleDefault(
        "project_reader_or_admin",
        "rule:project_reader_api or rule:admin_api",
        "Default rule for Project reader or admin APIs.",
        deprecated_rule=DEPRECATED_ADMIN_OR_OWNER)
]


def list_policies():
    return default_policies \
        + deprecated_default_policies



# Copyright (c) 2018 Huawei Technologies Co., Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from keystoneauth1 import exceptions as ks_exc
from oslo_config import cfg
from oslo_utils import uuidutils

from cyborg.common import exception as c_exc
from cyborg.common import placement_client
from cyborg.tests import base


class PlacementAPIClientTestCase(base.DietTestCase):
    """Test the Placement API client."""

    def setUp(self):
        super(PlacementAPIClientTestCase, self).setUp()
        self.mock_load_auth_p = mock.patch(
            'keystoneauth1.loading.load_auth_from_conf_options')
        self.mock_load_auth = self.mock_load_auth_p.start()
        self.mock_request_p = mock.patch(
            'keystoneauth1.session.Session.request')
        self.mock_request = self.mock_request_p.start()
        self.client = placement_client.PlacementClient()

    @mock.patch('keystoneauth1.session.Session')
    @mock.patch('keystoneauth1.loading.load_auth_from_conf_options')
    def test_constructor(self, load_auth_mock, ks_sess_mock):

        placement_client.PlacementClient()
        load_auth_mock.assert_called_once_with(cfg.CONF, 'placement')
        ks_sess_mock.assert_called_once_with(auth=load_auth_mock.return_value,
                                             cert=None,
                                             collect_timing=False,
                                             split_loggers=False,
                                             timeout=None,
                                             verify=True)

    @mock.patch('cyborg.common.placement_client.PlacementClient.post')
    def test_create_resource_provider(self, mock_post):
        rp_uuid = uuidutils.generate_uuid()
        self.client._create_resource_provider(self.context, rp_uuid, 'test')
        expected_url = '/resource_providers'
        mock_post.assert_called_once_with(
            expected_url, {'uuid': rp_uuid, 'name': 'test'},
            version='1.20', global_request_id=mock.ANY)

    @mock.patch('cyborg.common.placement_client.PlacementClient.delete')
    def test_delete_resource_provider(self, mock_delete):
        rp_uuid = uuidutils.generate_uuid()
        self.client.delete_provider(rp_uuid)
        expected_url = '/resource_providers/' + rp_uuid
        mock_delete.assert_called_once_with(
            expected_url, global_request_id=mock.ANY)

    def test_create_inventory(self):
        expected_payload = 'fake_inventory'
        rp_uuid = uuidutils.generate_uuid()
        e_filter = {'region_name': mock.ANY, 'service_type': 'placement'}
        self.client.create_inventory(rp_uuid, expected_payload)
        expected_url = '/resource_providers/%s/inventories' % rp_uuid
        self.mock_request.assert_called_once_with(expected_url, 'POST',
                                                  endpoint_filter=e_filter,
                                                  json=expected_payload)

    def test_get_inventory(self):
        rp_uuid = uuidutils.generate_uuid()
        e_filter = {'region_name': mock.ANY, 'service_type': 'placement'}
        resource_class = 'fake_resource_class'
        self.client.get_inventory(rp_uuid, resource_class)
        expected_url = '/resource_providers/%s/inventories/%s' % (
            rp_uuid, resource_class)
        self.mock_request.assert_called_once_with(expected_url, 'GET',
                                                  endpoint_filter=e_filter)

    def _test_get_inventory_not_found(self, details, expected_exception):
        rp_uuid = uuidutils.generate_uuid()
        resource_class = 'fake_resource_class'
        self.mock_request.side_effect = ks_exc.NotFound(details=details)
        self.assertRaises(expected_exception, self.client.get_inventory,
                          rp_uuid, resource_class)

    def test_get_inventory_not_found_no_resource_provider(self):
        self._test_get_inventory_not_found(
            "No resource provider with uuid",
            c_exc.PlacementResourceProviderNotFound)

    def test_get_inventory_not_found_no_inventory(self):
        self._test_get_inventory_not_found(
            "No inventory of class", c_exc.PlacementInventoryNotFound)

    def test_get_inventory_not_found_unknown_cause(self):
        self._test_get_inventory_not_found("Unknown cause", ks_exc.NotFound)

    def test_update_inventory(self):
        expected_payload = 'fake_inventory'
        rp_uuid = uuidutils.generate_uuid()
        e_filter = {'region_name': mock.ANY, 'service_type': 'placement'}
        resource_class = 'fake_resource_class'
        self.client.update_inventory(rp_uuid, expected_payload, resource_class)
        expected_url = '/resource_providers/%s/inventories/%s' % (
            rp_uuid, resource_class)
        self.mock_request.assert_called_once_with(expected_url, 'PUT',
                                                  endpoint_filter=e_filter,
                                                  json=expected_payload)

    def test_update_inventory_conflict(self):
        rp_uuid = uuidutils.generate_uuid()
        expected_payload = 'fake_inventory'
        resource_class = 'fake_resource_class'
        self.mock_request.side_effect = ks_exc.Conflict
        self.assertRaises(c_exc.PlacementInventoryUpdateConflict,
                          self.client.update_inventory, rp_uuid,
                          expected_payload, resource_class)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cyborg agent rpcapi test cases."""

import oslo_messaging as messaging

from cyborg.agent.rpcapi import AgentAPI
from cyborg.common import constants
from cyborg.common import rpc
from cyborg.objects import base as objects_base
from cyborg.tests import base
from oslo_context import context as oslo_context
from unittest import mock


class TestRPCAPI(base.TestCase):
    """Test agent rpcapi"""

    RPC_API_VERSION = '1.1'

    def setUp(self, topic=None):
        super(TestRPCAPI, self).setUp()
        self.topic = topic or constants.AGENT_TOPIC
        target = messaging.Target(topic=self.topic,
                                  version=self.RPC_API_VERSION)
        self.agent_rpcapi = AgentAPI()
        self.serializer = objects_base.CyborgObjectSerializer()
        self.client = rpc.get_client(target,
                                     version_cap=self.RPC_API_VERSION,
                                     serializer=self.serializer)

    def _test_rpc_call(self, method):
        ctxt = oslo_context.RequestContext(user_id='fake_user',
                                           project_id='fake_project')
        expect_val = True
        with mock.patch.object(self.agent_rpcapi,
                               'fpga_program') as mock_program:
            func_obj = getattr(self.agent_rpcapi, method)
            mock_program.return_value = expect_val
            actual_val = func_obj(ctxt, 'fake_dep_uuid')
            self.assertEqual(actual_val, expect_val)

    def test_fpga_program(self):
        self._test_rpc_call('fpga_program')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils

LOG = logging.getLogger(__name__)


class VhostTgt(object):

    def __init__(self, py):
        super(VhostTgt, self).__init__()
        self.py = py

    def get_rpc_methods(self):
        rpc_methods = self._get_json_objs('get_rpc_methods', '127.0.0.1')
        return rpc_methods

    def get_scsi_devices(self):
        scsi_devices = self._get_json_objs(
            'get_scsi_devices', '127.0.0.1')
        return scsi_devices

    def get_luns(self):
        luns = self._get_json_objs('get_luns', '127.0.0.1')
        return luns

    def get_interfaces(self):
        interfaces = self._get_json_objs(
            'get_interfaces', '127.0.0.1')
        return interfaces

    def add_ip_address(self, ifc_index, ip_addr):
        sub_args = [ifc_index, ip_addr]
        res = self.py.exec_rpc(
            'add_ip_address',
            '127.0.0.1',
            sub_args=sub_args)
        return res

    def delete_ip_address(self, ifc_index, ip_addr):
        sub_args = [ifc_index, ip_addr]
        res = self.py.exec_rpc(
            'delete_ip_address',
            '127.0.0.1',
            sub_args=sub_args)
        return res

    def get_bdevs(self):
        block_devices = self._get_json_objs(
            'get_bdevs', '127.0.0.1')
        return block_devices

    def delete_bdev(self, name):
        sub_args = [name]
        res = self.py.exec_rpc('delete_bdev', '127.0.0.1', sub_args=sub_args)
        LOG.info(res)

    def kill_instance(self, sig_name):
        sub_args = [sig_name]
        res = self.py.exec_rpc('kill_instance', '127.0.0.1', sub_args=sub_args)
        LOG.info(res)

    def construct_aio_bdev(self, filename, name, block_size):
        sub_args = [filename, name, str(block_size)]
        res = self.py.exec_rpc(
            'construct_aio_bdev',
            '127.0.0.1',
            sub_args=sub_args)
        LOG.info(res)

    def construct_error_bdev(self, basename):
        sub_args = [basename]
        res = self.py.exec_rpc(
            'construct_error_bdev',
            '127.0.0.1',
            sub_args=sub_args)
        LOG.info(res)

    def construct_nvme_bdev(
            self,
            name,
            trtype,
            traddr,
            adrfam=None,
            trsvcid=None,
            subnqn=None):
        sub_args = ["-b", "-t", "-a"]
        sub_args.insert(1, name)
        sub_args.insert(2, trtype)
        sub_args.insert(3, traddr)
        if adrfam is not None:
            sub_args.append("-f")
            sub_args.append(adrfam)
        if trsvcid is not None:
            sub_args.append("-s")
            sub_args.append(trsvcid)
        if subnqn is not None:
            sub_args.append("-n")
            sub_args.append(subnqn)
        res = self.py.exec_rpc(
            'construct_nvme_bdev',
            '127.0.0.1',
            sub_args=sub_args)
        return res

    def construct_null_bdev(self, name, total_size, block_size):
        sub_args = [name, str(total_size), str(block_size)]
        res = self.py.exec_rpc(
            'construct_null_bdev',
            '127.0.0.1',
            sub_args=sub_args)
        return res

    def construct_malloc_bdev(self, total_size, block_size):
        sub_args = [str(total_size), str(block_size)]
        res = self.py.exec_rpc(
            'construct_malloc_bdev',
            '10.0.2.15',
            sub_args=sub_args)
        LOG.info(res)

    def _get_json_objs(self, method, server_ip):
        res = self.py.exec_rpc(method, server_ip)
        json_obj = jsonutils.loads(res)
        return json_obj



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#  cyborg documentation build configuration file
#
# Refer to the Sphinx documentation for advice on configuring this file:
#
#   http://www.sphinx-doc.org/en/stable/config.html

import os
import sys

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.graphviz',
    'openstackdocstheme',
    'oslo_config.sphinxconfiggen',
    'oslo_config.sphinxext',
    'oslo_policy.sphinxext',
    'oslo_policy.sphinxpolicygen',
    'sphinxcontrib.rsvgconverter',
]

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = '2013, OpenStack Foundation'

# openstackdocstheme options
openstackdocs_auto_name = False
openstackdocs_repo_name = 'openstack/cyborg'
openstackdocs_bug_project = 'cyborg'
openstackdocs_bug_tag = 'doc'
openstackdocs_pdf_link = True

config_generator_config_file = \
    '../../tools/config/cyborg-config-generator.conf'
sample_config_basename = '_static/cyborg'

policy_generator_config_file = [
    ('../../tools/config/cyborg-policy-generator.conf',
     '_static/cyborg'),
]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'cyborgdoc'

# The name for this set of Sphinx documents. If None, it defaults to
# "<project> v<release> documentation".
html_title = 'Cyborg'
html_theme = 'openstackdocs'

# -- Options for LaTeX output -------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # openany: Skip blank pages in generated PDFs
    'maxlistdepth': 10,
    'extraclassoptions': 'openany,oneside',
    'preamble': r'\setcounter{tocdepth}{2}',
    'makeindex': '',
    'printindex': '',
}

# Disable usage of xindy https://bugzilla.redhat.com/show_bug.cgi?id=1643664
# Some distros are missing xindy
latex_use_xindy = False

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index',
     'doc-cyborg.tex',
     'Cyborg Documentation',
     'OpenStack Foundation', 'manual'),
]

# Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {'https://docs.python.org/': None}



# Copyright 2017 Huawei Technologies Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""SQLAlchemy storage backend."""

import copy
import threading
import uuid

from oslo_db import api as oslo_db_api
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_log import log
from oslo_utils import strutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import load_only

from cyborg.common import exception
from cyborg.common.i18n import _
from cyborg.db import api
from cyborg.db.sqlalchemy import models

_CONTEXT = threading.local()
LOG = log.getLogger(__name__)

main_context_manager = enginefacade.transaction_context()


def get_backend():
    """The backend is this module itself."""
    return Connection()


def _session_for_read():
    return enginefacade.reader.using(_CONTEXT)


def _session_for_write():
    return enginefacade.writer.using(_CONTEXT)


def get_session(use_slave=False, **kwargs):
    return main_context_manager._factory.get_legacy_facade().get_session(
        use_slave=use_slave, **kwargs)


def model_query(context, model, *args, **kwargs):
    """Query helper for simpler session usage.

    :param context: Context of the query
    :param model: Model to query. Must be a subclass of ModelBase.
    :param args: Arguments to query. If None - model is used.

    Keyword arguments:

    :keyword project_only:
      If set to True, then will do query filter with context's project_id.
      if set to False or absent, then will not do query filter with context's
      project_id.
    :type project_only: bool
    """

    if kwargs.pop("project_only", False):
        kwargs["project_id"] = context.project_id

    with _session_for_read() as session:
        query = sqlalchemyutils.model_query(
            model, session, args, **kwargs)
        return query


def add_identity_filter(query, value):
    """Adds an identity filter to a query.

    Filters results by ID, if supplied value is a valid integer.
    Otherwise attempts to filter results by UUID.

    :param query: Initial query to add filter to.
    :param value: Value for filtering results by.
    :return: Modified query.
    """
    if strutils.is_int_like(value):
        return query.filter_by(id=value)
    elif uuidutils.is_uuid_like(value):
        return query.filter_by(uuid=value)
    else:
        raise exception.InvalidIdentity(identity=value)


def _paginate_query(context, model, query, limit=None, marker=None,
                    sort_key=None, sort_dir=None):
    sort_keys = ['id']
    if sort_key and sort_key not in sort_keys:
        sort_keys.insert(0, sort_key)
    try:
        query = sqlalchemyutils.paginate_query(query, model, limit, sort_keys,
                                               marker=marker,
                                               sort_dir=sort_dir)
    except db_exc.InvalidSortKey:
        raise exception.InvalidParameterValue(
            _('The sort_key value "%(key)s" is an invalid field for sorting')
            % {'key': sort_key})
    return query.all()


class Connection(api.Connection):
    """SqlAlchemy connection."""

    def __init__(self):
        pass

    def attach_handle_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()

        attach_handle = models.AttachHandle()
        attach_handle.update(values)

        with _session_for_write() as session:
            try:
                session.add(attach_handle)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.AttachHandleAlreadyExists(uuid=values['uuid'])
            return attach_handle

    def attach_handle_get_by_uuid(self, context, uuid):
        query = model_query(
            context,
            models.AttachHandle).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='AttachHandle',
                msg='with uuid=%s' % uuid)

    def attach_handle_get_by_id(self, context, id):
        query = model_query(
            context,
            models.AttachHandle).filter_by(id=id)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='AttachHandle',
                msg='with id=%s' % id)

    def attach_handle_list_by_type(self, context, attach_type='PCI'):
        query = model_query(context, models.AttachHandle). \
            filter_by(attach_type=attach_type)
        try:
            return query.all()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='AttachHandle',
                msg='with type=%s' % attach_type)

    def attach_handle_get_by_filters(self, context,
                                     filters, sort_key='created_at',
                                     sort_dir='desc', limit=None,
                                     marker=None, join_columns=None):
        """Return attach_handle that match all filters sorted by the given
        keys. Deleted attach_handle will be returned by default, unless
        there's a filter that says otherwise.
        """

        if limit == 0:
            return []

        query_prefix = model_query(context, models.AttachHandle)
        filters = copy.deepcopy(filters)

        exact_match_filter_names = ['uuid', 'id', 'deployable_id', 'cpid_id']

        # Filter the query
        query_prefix = self._exact_filter(models.AttachHandle, query_prefix,
                                          filters, exact_match_filter_names)
        if query_prefix is None:
            return []
        return _paginate_query(context, models.AttachHandle, query_prefix,
                               limit, marker, sort_key, sort_dir)

    def _exact_filter(self, model, query, filters, legal_keys=None):
        """Applies exact match filtering to a deployable query.
        Returns the updated query.  Modifies filters argument to remove
        filters consumed.
        :param model: DB model
        :param query: query to apply filters to
        :param filters: dictionary of filters; values that are lists,
                        tuples, sets, or frozensets cause an 'IN' test to
                        be performed, while exact matching ('==' operator)
                        is used for other values
        :param legal_keys: list of keys to apply exact filtering to
        """

        if filters is None:
            filters = {}
        if legal_keys is None:
            legal_keys = []
        filter_dict = {}

        # Walk through all the keys
        for key in legal_keys:
            # Skip ones we're not filtering on
            if key not in filters:
                continue

            # OK, filtering on this key; what value do we search for?
            value = filters.pop(key)

            if isinstance(value, (list, tuple, set, frozenset)):
                if not value:
                    return None
                # Looking for values in a list; apply to query directly
                column_attr = getattr(model, key)
                query = query.filter(column_attr.in_(value))
            else:
                filter_dict[key] = value
        # Apply simple exact matches
        if filter_dict:
            query = query.filter(*[getattr(model, k) == v
                                   for k, v in filter_dict.items()])
        return query

    def attach_handle_list(self, context):
        query = model_query(context, models.AttachHandle)
        return _paginate_query(context, models.AttachHandle, query=query)

    def attach_handle_update(self, context, uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing AttachHandle.")
            raise exception.InvalidParameterValue(err=msg)
        return self._do_update_attach_handle(context, uuid, values)

    @oslo_db_api.retry_on_deadlock
    def _do_update_attach_handle(self, context, uuid, values):
        with _session_for_write():
            query = model_query(context, models.AttachHandle)
            query = add_identity_filter(query, uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='AttachHandle',
                    msg='with uuid=%s' % uuid)
            ref.update(values)
        return ref

    @oslo_db_api.retry_on_deadlock
    def _do_allocate_attach_handle(self, context, deployable_id):
        """Atomically get a set of attach handles that match the query
           and mark one of those as in_use.
        """
        with _session_for_write() as session:
            query = model_query(context, models.AttachHandle). \
                filter_by(deployable_id=deployable_id,
                          in_use=False)
            values = {"in_use": True}
            ref = query.with_for_update().first()
            if not ref:
                msg = 'Matching deployable_id {0}'.format(deployable_id)
                raise exception.ResourceNotFound(
                    resource='AttachHandle', msg=msg)
            ref.update(values)
            session.flush()
        return ref

    def attach_handle_allocate(self, context, deployable_id):
        """Allocate an attach handle with given deployable.

           To allocate is to get an unused resource and mark it as in_use.
        """
        try:
            ah = self._do_allocate_attach_handle(
                context, deployable_id)
        except NoResultFound:
            msg = 'Matching deployable_id {0}'.format(deployable_id)
            raise exception.ResourceNotFound(
                resource='AttachHandle', msg=msg)
        return ah

    # NOTE: For deallocate, we use attach_handle_update()

    @oslo_db_api.retry_on_deadlock
    def attach_handle_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.AttachHandle)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='AttachHandle',
                    msg='with uuid=%s' % uuid)

    def control_path_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()

        control_path_id = models.ControlpathID()
        control_path_id.update(values)

        with _session_for_write() as session:
            try:
                session.add(control_path_id)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.ControlpathIDAlreadyExists(uuid=values['uuid'])
            return control_path_id

    def control_path_get_by_uuid(self, context, uuid):
        query = model_query(
            context,
            models.ControlpathID).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='ControlpathID',
                msg='with uuid=%s' % uuid)

    def control_path_get_by_filters(self, context,
                                    filters, sort_key='created_at',
                                    sort_dir='desc', limit=None,
                                    marker=None, join_columns=None):
        """Return attach_handle that match all filters sorted by the given
        keys. Deleted attach_handle will be returned by default, unless
        there's a filter that says otherwise.
        """

        if limit == 0:
            return []

        query_prefix = model_query(context, models.ControlpathID)
        filters = copy.deepcopy(filters)

        exact_match_filter_names = ['uuid', 'id', 'device_id', 'cpid_info',
                                    'cpid_type']

        # Filter the query
        query_prefix = self._exact_filter(models.ControlpathID, query_prefix,
                                          filters, exact_match_filter_names)
        if query_prefix is None:
            return []
        return _paginate_query(context, models.ControlpathID, query_prefix,
                               limit, marker, sort_key, sort_dir)

    def control_path_list(self, context):
        query = model_query(context, models.ControlpathID)
        return _paginate_query(context, models.ControlpathID, query=query)

    def control_path_update(self, context, uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing ControlpathID.")
            raise exception.InvalidParameterValue(err=msg)
        return self._do_update_control_path(context, uuid, values)

    @oslo_db_api.retry_on_deadlock
    def _do_update_control_path(self, context, uuid, values):
        with _session_for_write():
            query = model_query(context, models.ControlpathID)
            query = add_identity_filter(query, uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='ControlpathID',
                    msg='with uuid=%s' % uuid)
            ref.update(values)
        return ref

    @oslo_db_api.retry_on_deadlock
    def control_path_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.ControlpathID)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ControlpathNotFound(uuid=uuid)

    def device_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()

        device = models.Device()
        device.update(values)

        with _session_for_write() as session:
            try:
                session.add(device)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.DeviceAlreadyExists(uuid=values['uuid'])
            return device

    def device_get(self, context, uuid):
        query = model_query(
            context,
            models.Device).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Device',
                msg='with uuid=%s' % uuid)

    def device_get_by_id(self, context, id):
        query = model_query(
            context,
            models.Device).filter_by(id=id)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Device',
                msg='with id=%s' % id)

    def device_list_by_filters(self, context,
                               filters, sort_key='created_at',
                               sort_dir='desc', limit=None,
                               marker=None, join_columns=None):
        """Return devices that match all filters sorted by the given keys."""

        if limit == 0:
            return []

        query_prefix = model_query(context, models.Device)
        filters = copy.deepcopy(filters)

        exact_match_filter_names = ['uuid', 'id', 'type',
                                    'vendor', 'model', 'hostname']

        # Filter the query
        query_prefix = self._exact_filter(models.Device, query_prefix,
                                          filters, exact_match_filter_names)
        if query_prefix is None:
            return []
        return _paginate_query(context, models.Device, query_prefix,
                               limit, marker, sort_key, sort_dir)

    def device_list(self, context, limit=None, marker=None, sort_key=None,
                    sort_dir=None):
        query = model_query(context, models.Device)
        return _paginate_query(context, models.Device, query,
                               limit, marker, sort_key, sort_dir)

    def device_update(self, context, uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing Device.")
            raise exception.InvalidParameterValue(err=msg)

        try:
            return self._do_update_device(context, uuid, values)
        except db_exc.DBDuplicateEntry as e:
            if 'name' in e.columns:
                raise exception.DuplicateDeviceName(name=values['name'])

    @oslo_db_api.retry_on_deadlock
    def _do_update_device(self, context, uuid, values):
        with _session_for_write():
            query = model_query(context, models.Device)
            query = add_identity_filter(query, uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='Device',
                    msg='with uuid=%s' % uuid)

            ref.update(values)
        return ref

    @oslo_db_api.retry_on_deadlock
    def device_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.Device)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='Device',
                    msg='with uuid=%s' % uuid)

    def device_profile_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()

        device_profile = models.DeviceProfile()
        device_profile.update(values)

        with _session_for_write() as session:
            try:
                session.add(device_profile)
                session.flush()
            except db_exc.DBDuplicateEntry as e:
                # mysql duplicate key error changed as reference link below:
                # https://review.opendev.org/c/openstack/oslo.db/+/792124
                LOG.info('Duplicate columns are: ', e.columns)
                columns = [column.split('0')[1] if 'uniq_' in column else
                           column for column in e.columns]
                if 'name' in columns:
                    raise exception.DuplicateDeviceProfileName(
                        name=values['name'])
                else:
                    raise exception.DeviceProfileAlreadyExists(
                        uuid=values['uuid'])
            return device_profile

    def device_profile_get_by_uuid(self, context, uuid):
        query = model_query(
            context,
            models.DeviceProfile).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Device Profile',
                msg='with uuid=%s' % uuid)

    def device_profile_get_by_id(self, context, id):
        query = model_query(
            context,
            models.DeviceProfile).filter_by(id=id)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Device Profile',
                msg='with id=%s' % id)

    def device_profile_get(self, context, name):
        query = model_query(
            context, models.DeviceProfile).filter_by(name=name)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Device Profile',
                msg='with name=%s' % name)

    def device_profile_list_by_filters(
            self, context, filters, sort_key='created_at', sort_dir='desc',
            limit=None, marker=None, join_columns=None):
        if limit == 0:
            return []

        query_prefix = model_query(context, models.DeviceProfile)
        filters = copy.deepcopy(filters)

        exact_match_filter_names = ['uuid', 'id', 'name']

        # Filter the query
        query_prefix = self._exact_filter(models.DeviceProfile, query_prefix,
                                          filters, exact_match_filter_names)
        if query_prefix is None:
            return []
        return _paginate_query(context, models.DeviceProfile, query_prefix,
                               limit, marker, sort_key, sort_dir)

    def device_profile_list(self, context):
        query = model_query(context, models.DeviceProfile)
        return _paginate_query(context, models.DeviceProfile, query=query)

    def device_profile_update(self, context, uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing DeviceProfile.")
            raise exception.InvalidParameterValue(err=msg)

        try:
            return self._do_update_device_profile(context, uuid, values)
        except db_exc.DBDuplicateEntry as e:
            if 'name' in e.columns:
                raise exception.DuplicateDeviceProfileName(name=values['name'])

    @oslo_db_api.retry_on_deadlock
    def _do_update_device_profile(self, context, uuid, values):
        with _session_for_write():
            query = model_query(context, models.DeviceProfile)
            query = add_identity_filter(query, uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='Device Profile',
                    msg='with uuid=%s' % uuid)

            ref.update(values)
        return ref

    @oslo_db_api.retry_on_deadlock
    def device_profile_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.DeviceProfile)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='Device Profile',
                    msg='with uuid=%s' % uuid)

    def deployable_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()
        if values.get('id'):
            values.pop('id', None)
        deployable = models.Deployable()
        deployable.update(values)

        with _session_for_write() as session:
            try:
                session.add(deployable)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.DeployableAlreadyExists(uuid=values['uuid'])
            return deployable

    def deployable_get(self, context, uuid):
        query = model_query(
            context,
            models.Deployable).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Deployable',
                msg='with uuid=%s' % uuid)

    def deployable_get_by_rp_uuid(self, context, rp_uuid):
        """Get a deployable by resource provider UUID."""
        query = model_query(
            context,
            models.Deployable).filter_by(rp_uuid=rp_uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Deployable',
                msg='with resource provider uuid=%s' % rp_uuid)

    def deployable_list(self, context):
        query = model_query(context, models.Deployable)
        return query.all()

    def deployable_update(self, context, uuid, values):
        if 'uuid' in values:
            msg = _("Cannot overwrite UUID for an existing Deployable.")
            raise exception.InvalidParameterValue(err=msg)

        try:
            return self._do_update_deployable(context, uuid, values)
        except db_exc.DBDuplicateEntry as e:
            if 'name' in e.columns:
                raise exception.DuplicateDeployableName(name=values['name'])

    @oslo_db_api.retry_on_deadlock
    def _do_update_deployable(self, context, uuid, values):
        with _session_for_write():
            query = model_query(context, models.Deployable)
            # query = add_identity_filter(query, uuid)
            query = query.filter_by(uuid=uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='Deployable',
                    msg='with uuid=%s' % uuid)

            ref.update(values)
        return ref

    @oslo_db_api.retry_on_deadlock
    def deployable_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.Deployable)
            query = add_identity_filter(query, uuid)
            query.update({'root_id': None})
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='Deployable',
                    msg='with uuid=%s' % uuid)

    def deployable_get_by_filters(self, context,
                                  filters, sort_key='created_at',
                                  sort_dir='desc', limit=None,
                                  marker=None, join_columns=None):
        """Return list of deployables matching all filters sorted by
        the sort_key. See deployable_get_by_filters_sort for
        more information.
        """
        return self.deployable_get_by_filters_sort(context, filters,
                                                   limit=limit, marker=marker,
                                                   join_columns=join_columns,
                                                   sort_key=sort_key,
                                                   sort_dir=sort_dir)

    def deployable_get_by_filters_sort(self, context, filters, limit=None,
                                       marker=None, join_columns=None,
                                       sort_key=None, sort_dir=None):
        """Return deployables that match all filters sorted by the given
        keys. Deleted deployables will be returned by default, unless
        there's a filter that says otherwise.
        """
        if limit == 0:
            return []

        query_prefix = model_query(context, models.Deployable)
        filters = copy.deepcopy(filters)

        exact_match_filter_names = ['id', 'uuid', 'name',
                                    'parent_id', 'root_id',
                                    'num_accelerators', 'device_id',
                                    'driver_name', 'rp_uuid', 'bitstream_id']

        # Filter the query
        query_prefix = self._exact_filter(models.Deployable, query_prefix,
                                          filters,
                                          exact_match_filter_names)
        if query_prefix is None:
            return []
        return _paginate_query(context, models.Deployable, query_prefix,
                               limit, marker, sort_key, sort_dir)

    def attribute_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()
        if values.get('id'):
            values.pop('id', None)
        attribute = models.Attribute()
        attribute.update(values)

        with _session_for_write() as session:
            try:
                session.add(attribute)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.AttributeAlreadyExists(
                    uuid=values['uuid'])
            return attribute

    def attribute_get(self, context, uuid):
        query = model_query(
            context,
            models.Attribute).filter_by(uuid=uuid)
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='Attribute',
                msg='with uuid=%s' % uuid)

    def attribute_get_by_deployable_id(self, context, deployable_id):
        query = model_query(
            context,
            models.Attribute).filter_by(deployable_id=deployable_id)
        return query.all()

    def attribute_get_by_filter(self, context, filters):
        """Return attributes that matches the filters
        """
        query_prefix = model_query(context, models.Attribute)
        exact_match_filter_names = ['deployable_id', 'key']

        # Filter the query
        query_prefix = self._exact_filter(models.Attribute, query_prefix,
                                          filters, exact_match_filter_names)
        if query_prefix is None:
            return []

        return query_prefix.all()

    # def _exact_attribute_by_filter(self, query, filters):
    #     """Applies exact match filtering to a atrtribute query.
    #     Returns the updated query.
    #     :param filters: The filters specified by a dict of kv pairs
    #     """
    #
    #     model = models.Attribute
    #     filter_dict = filters
    #
    #     # Apply simple exact matches
    #     query = query.filter(*[getattr(models.Attribute, k) == v
    #                            for k, v in filter_dict.items()])
    #     return query

    def attribute_update(self, context, uuid, key, value):
        return self._do_update_attribute(context, uuid, key, value)

    @oslo_db_api.retry_on_deadlock
    def _do_update_attribute(self, context, uuid, key, value):
        update_fields = {'key': key, 'value': value}
        with _session_for_write():
            query = model_query(context, models.Attribute)
            query = add_identity_filter(query, uuid)
            try:
                ref = query.with_for_update().one()
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='Attribute',
                    msg='with uuid=%s' % uuid)

            ref.update(update_fields)
        return ref

    def attribute_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.Attribute)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='Attribute',
                    msg='with uuid=%s' % uuid)

    def extarq_create(self, context, values):
        if not values.get('uuid'):
            values['uuid'] = uuidutils.generate_uuid()
        if values.get('id'):
            values.pop('id', None)

        if values.get('device_profile_id'):
            pass  # Already have the devprof id, so nothing to do
        elif values.get('device_profile_name'):
            devprof = self.device_profile_get(context,
                                              values['device_profile_name'])
            values['device_profile_id'] = devprof['id']
        else:
            raise exception.DeviceProfileNameNeeded()

        extarq = models.ExtArq()
        extarq.update(values)

        with _session_for_write() as session:
            try:
                session.add(extarq)
                session.flush()
            except db_exc.DBDuplicateEntry:
                raise exception.ExtArqAlreadyExists(uuid=values['uuid'])
            return extarq

    @oslo_db_api.retry_on_deadlock
    def extarq_delete(self, context, uuid):
        with _session_for_write():
            query = model_query(context, models.ExtArq)
            query = add_identity_filter(query, uuid)
            count = query.delete()
            if count != 1:
                raise exception.ResourceNotFound(
                    resource='ExtArq',
                    msg='with uuid=%s' % uuid)

    def extarq_update(self, context, uuid, values, state_scope=None):
        if 'uuid' in values and values['uuid'] != uuid:
            msg = _("Cannot overwrite UUID for an existing ExtArq.")
            raise exception.InvalidParameterValue(err=msg)
        return self._do_update_extarq(context, uuid, values, state_scope)

    @oslo_db_api.retry_on_deadlock
    def _do_update_extarq(self, context, uuid, values, state_scope=None):
        with _session_for_write():
            query = model_query(context, models.ExtArq)
            query = query_update = query.filter_by(
                uuid=uuid).with_for_update()
            if type(state_scope) is list:
                query_update = query_update.filter(
                    models.ExtArq.state.in_(state_scope))
            try:
                query_update.update(
                    values, synchronize_session="fetch")
            except NoResultFound:
                raise exception.ResourceNotFound(
                    resource='ExtArq',
                    msg='with uuid=%s' % uuid)
            ref = query.first()
        return ref

    def extarq_list(self, context, uuid_range=None):
        query = model_query(context, models.ExtArq)
        if type(uuid_range) is list:
            query = query.filter(
                models.ExtArq.uuid.in_(uuid_range))
        return _paginate_query(context, models.ExtArq, query)

    @oslo_db_api.retry_on_deadlock
    def extarq_get(self, context, uuid, lock=False):
        query = model_query(
            context,
            models.ExtArq).filter_by(uuid=uuid)
        # NOTE we will support aync bind, so get query by lock
        if lock:
            query = query.with_for_update()
        try:
            return query.one()
        except NoResultFound:
            raise exception.ResourceNotFound(
                resource='ExtArq',
                msg='with uuid=%s' % uuid)

    def _get_quota_usages(self, context, project_id, resources=None):
        # Broken out for testability
        query = model_query(context, models.QuotaUsage,).filter_by(
            project_id=project_id)
        if resources:
            query = query.filter(models.QuotaUsage.resource.in_(
                list(resources)))
        rows = query.order_by(models.QuotaUsage.id.asc()). \
            with_for_update().all()
        return {row.resource: row for row in rows}

    def _quota_usage_create(self, project_id, resource, until_refresh,
                            in_use, reserved, session=None):

        quota_usage_ref = models.QuotaUsage()
        quota_usage_ref.project_id = project_id
        quota_usage_ref.resource = resource
        quota_usage_ref.in_use = in_use
        quota_usage_ref.reserved = reserved
        quota_usage_ref.until_refresh = until_refresh
        quota_usage_ref.save(session=session)

        return quota_usage_ref

    def _reservation_create(self, uuid, usage, project_id, resource, delta,
                            expire, session=None):
        usage_id = usage['id'] if usage else None
        reservation_ref = models.Reservation()
        reservation_ref.uuid = uuid
        reservation_ref.usage_id = usage_id
        reservation_ref.project_id = project_id
        reservation_ref.resource = resource
        reservation_ref.delta = delta
        reservation_ref.expire = expire
        reservation_ref.save(session=session)
        return reservation_ref

    def _get_reservation_resources(self, context, reservation_ids):
        """Return the relevant resources by reservations."""

        reservations = model_query(context, models.Reservation). \
            options(load_only('resource')). \
            filter(models.Reservation.uuid.in_(reservation_ids)). \
            all()
        return {r.resource for r in reservations}

    def _quota_reservations(self, session, context, reservations):
        """Return the relevant reservations."""

        # Get the listed reservations
        return model_query(context, models.Reservation). \
            filter(models.Reservation.uuid.in_(reservations)). \
            with_for_update(). \
            all()

    def quota_reserve(self, context, resources, deltas, expire,
                      until_refresh, max_age, project_id=None,
                      is_allocated_reserve=False):
        """Create reservation record in DB according to params"""
        with _session_for_write() as session:
            if project_id is None:
                project_id = context.project_id
            usages = self._get_quota_usages(context, project_id,
                                            resources=deltas.keys())
            work = set(deltas.keys())
            while work:
                resource = work.pop()

                # Do we need to refresh the usage?
                refresh = False
                # create quota usage in DB if there is no record of this type
                # of resource
                if resource not in usages:
                    usages[resource] = self._quota_usage_create(
                        project_id, resource, until_refresh or None,
                        in_use=0, reserved=0, session=session)
                    refresh = True
                elif usages[resource].in_use < 0:
                    # Negative in_use count indicates a desync, so try to
                    # heal from that...
                    refresh = True
                elif usages[resource].until_refresh is not None:
                    usages[resource].until_refresh -= 1
                    if usages[resource].until_refresh <= 0:
                        refresh = True
                elif max_age and usages[resource].updated_at is not None and (
                    (timeutils.utcnow() -
                        usages[resource].updated_at).total_seconds() >=
                        max_age):
                    refresh = True

                # refresh the usage
                if refresh:
                    # Grab the sync routine
                    updates = self._sync_acc_res(context,
                                                 resource, project_id)
                    for res, in_use in updates.items():
                        # Make sure we have a destination for the usage!
                        if res not in usages:
                            usages[res] = self._quota_usage_create(
                                project_id,
                                res,
                                until_refresh or None,
                                in_use=0,
                                reserved=0,
                                session=session
                            )

                        # Update the usage
                        usages[res].in_use = in_use
                        usages[res].until_refresh = until_refresh or None

                        # Because more than one resource may be refreshed
                        # by the call to the sync routine, and we don't
                        # want to double-sync, we make sure all refreshed
                        # resources are dropped from the work set.
                        work.discard(res)

                        # NOTE(Vek): We make the assumption that the sync
                        #            routine actually refreshes the
                        #            resources that it is the sync routine
                        #            for.  We don't check, because this is
                        #            a best-effort mechanism.

            unders = [r for r, delta in deltas.items()
                      if delta < 0 and delta + usages[r].in_use < 0]
            reservations = []
            for resource, delta in deltas.items():
                usage = usages[resource]
                reservation = self._reservation_create(
                    str(uuid.uuid4()), usage, project_id, resource,
                    delta, expire, session=session)
                reservations.append(reservation.uuid)
                usages[resource].reserved += delta
            session.flush()
        if unders:
            LOG.warning("Change will make usage less than 0 for the "
                        "following resources: %s", unders)
        return reservations

    def _sync_acc_res(self, context, resource, project_id):
        """Quota sync funciton"""
        res_in_use = self._device_data_get_for_project(context, resource,
                                                       project_id)
        return {resource: res_in_use}

    def _device_data_get_for_project(self, context, resource, project_id):
        """Return the number of resource which is being used by a project"""
        query = model_query(context, models.Device).filter_by(type=resource)

        return query.count()

    def _dict_with_usage_id(self, usages):
        return {row.id: row for row in usages.values()}

    def reservation_commit(self, context, reservations, project_id=None):
        """Commit quota reservation to quota usage table"""
        with _session_for_write() as session:
            quota_usage = self._get_quota_usages(
                context, project_id,
                resources=self._get_reservation_resources(context,
                                                          reservations))
            usages = self._dict_with_usage_id(quota_usage)

            for reservation in self._quota_reservations(session, context,
                                                        reservations):

                usage = usages[reservation.usage_id]
                if reservation.delta >= 0:
                    usage.reserved -= reservation.delta
                usage.in_use += reservation.delta
                session.flush()
                reservation.delete(session=session)

    def process_sort_params(self, sort_keys, sort_dirs,
                            default_keys=['created_at', 'id'],
                            default_dir='asc'):

        # Determine direction to use for when adding default keys
        if sort_dirs and len(sort_dirs) != 0:
            default_dir_value = sort_dirs[0]
        else:
            default_dir_value = default_dir

        # Create list of keys (do not modify the input list)
        if sort_keys:
            result_keys = list(sort_keys)
        else:
            result_keys = []

        # If a list of directions is not provided,
        # use the default sort direction for all provided keys
        if sort_dirs:
            result_dirs = []
            # Verify sort direction
            for sort_dir in sort_dirs:
                if sort_dir not in ('asc', 'desc'):
                    msg = _("Unknown sort direction, must be 'desc' or 'asc'")
                    raise exception.InvalidInput(reason=msg)
                result_dirs.append(sort_dir)
        else:
            result_dirs = [default_dir_value for _sort_key in result_keys]

        # Ensure that the key and direction length match
        while len(result_dirs) < len(result_keys):
            result_dirs.append(default_dir_value)
        # Unless more direction are specified, which is an error
        if len(result_dirs) > len(result_keys):
            msg = _("Sort direction size exceeds sort key size")
            raise exception.InvalidInput(reason=msg)

        # Ensure defaults are included
        for key in default_keys:
            if key not in result_keys:
                result_keys.append(key)
                result_dirs.append(default_dir_value)

        return result_keys, result_dirs



# Copyright 2019 Intel, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cyborg.tests.unit.api.controllers.v2 import base as v2_test
from cyborg.tests.unit import fake_device


class TestDevicesController(v2_test.APITestV2):

    DEVICE_URL = '/devices'

    def setUp(self):
        super(TestDevicesController, self).setUp()
        self.headers = self.gen_headers(self.context)
        self.fake_devices = fake_device.get_fake_devices_objs()

    def _validate_links(self, links, device_uuid):
        has_self_link = False
        for link in links:
            if link['rel'] == 'self':
                has_self_link = True
                url = link['href']
                components = url.split('/')
                self.assertEqual(components[-1], device_uuid)
        self.assertTrue(has_self_link)

    def _validate_device(self, in_device, out_device):
        for field in in_device.keys():
            if field != 'id':
                self.assertEqual(in_device[field], out_device[field])
        # Check that the link is properly set up
        self._validate_links(out_device['links'], in_device['uuid'])

    @mock.patch('cyborg.objects.Device.get')
    def test_get_one_by_uuid(self, mock_device):
        in_device = self.fake_devices[0]
        mock_device.return_value = in_device
        uuid = in_device['uuid']

        url = self.DEVICE_URL + '/%s'
        out_device = self.get_json(url % uuid, headers=self.headers)
        mock_device.assert_called_once()
        self._validate_device(in_device, out_device)

    @mock.patch('cyborg.objects.Device.list')
    def test_get_all(self, mock_devices):
        mock_devices.return_value = self.fake_devices
        data = self.get_json(self.DEVICE_URL, headers=self.headers)
        out_devices = data['devices']
        self.assertIsInstance(out_devices, list)
        for out_dev in out_devices:
            self.assertIsInstance(out_dev, dict)
        self.assertEqual(len(out_devices), len(self.fake_devices))
        for in_device, out_device in zip(self.fake_devices, out_devices):
            self._validate_device(in_device, out_device)

    @mock.patch('cyborg.objects.Device.list')
    def test_get_with_filters(self, mock_devices):
        in_devices = self.fake_devices
        mock_devices.return_value = in_devices[:1]
        data = self.get_json(
            self.DEVICE_URL + "?filters.field=limit&filters.value=1",
            headers=self.headers)
        out_devices = data['devices']
        mock_devices.assert_called_once_with(mock.ANY, filters={"limit": "1"})
        for in_device, out_device in zip(self.fake_devices, out_devices):
            self._validate_device(in_device, out_device)

    @mock.patch('cyborg.objects.Device.list')
    def test_get_by_type(self, mock_devices):
        in_devices = self.fake_devices
        mock_devices.return_value = [in_devices[0]]
        data = self.get_json(
            self.DEVICE_URL + "?type=FPGA",
            headers=self.headers)
        out_devices = data['devices']
        mock_devices.assert_called_once_with(mock.ANY,
                                             filters={"type": "FPGA"})
        for in_device, out_device in zip(self.fake_devices, out_devices):
            self._validate_device(in_device, out_device)

    @mock.patch('cyborg.objects.Device.list')
    def test_get_by_vendor(self, mock_devices):
        in_devices = self.fake_devices
        mock_devices.return_value = [in_devices[0]]
        data = self.get_json(
            self.DEVICE_URL + "?vendor=0xABCD",
            headers=self.headers)
        out_devices = data['devices']
        mock_devices.assert_called_once_with(mock.ANY,
                                             filters={"vendor": "0xABCD"})
        for in_device, out_device in zip(self.fake_devices, out_devices):
            self._validate_device(in_device, out_device)

    @mock.patch('cyborg.objects.Device.list')
    def test_get_by_hostname(self, mock_devices):
        in_devices = self.fake_devices
        mock_devices.return_value = [in_devices[0]]
        data = self.get_json(
            self.DEVICE_URL + "?hostname=test-node-1",
            headers=self.headers)
        out_devices = data['devices']
        mock_devices.assert_called_once_with(
            mock.ANY, filters={"hostname": "test-node-1"})
        for in_device, out_device in zip(self.fake_devices, out_devices):
            self._validate_device(in_device, out_device)



# Copyright 2020 Inspur Electronic Information Industry Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_serialization import jsonutils

from cyborg.accelerator.drivers.ssd.base import SSDDriver
from cyborg.accelerator.drivers.ssd.inspur.driver import InspurNVMeSSDDriver
from cyborg.accelerator.drivers.ssd import utils
from cyborg.tests import base

NVME_SSD_INFO = \
    "0000:db:00.0 Non-Volatile memory controller [0108]: Inspur " \
    "Electronic Information Industry Co., Ltd. Device [1bd4:1001]" \
    " (rev 02)\n0000:db:01.0 Non-Volatile memory controller " \
    "[0108]: Inspur Electronic Information Industry Co., Ltd. " \
    "Device [1bd4:1001] (rev 02)"


class stdout(object):
    def readlines(self):
        return [NVME_SSD_INFO]


class p(object):
    def __init__(self):
        self.stdout = stdout()

    def wait(self):
        pass


class TestSSDDriverUtils(base.TestCase):

    def setUp(self):
        super().setUp()
        self.p = p()

    @mock.patch('cyborg.accelerator.drivers.ssd.utils.lspci_privileged')
    def test_discover_vendors(self, mock_devices):
        mock_devices.return_value = self.p.stdout.readlines()
        ssd_vendors = utils.discover_vendors()
        self.assertEqual(1, len(ssd_vendors))

    @mock.patch('cyborg.accelerator.drivers.ssd.utils.lspci_privileged')
    def test_discover_with_inspur_ssd_driver(self, mock_devices_for_vendor):
        mock_devices_for_vendor.return_value = self.p.stdout.readlines()
        self.set_defaults(host='host-192-168-32-195', debug=True)
        vendor_id = '1bd4'
        ssd_list = InspurNVMeSSDDriver.discover(vendor_id)
        self.assertEqual(2, len(ssd_list))
        attach_handle_list = [
            {'attach_type': 'PCI',
             'attach_info': '{"bus": "db", '
                            '"device": "00", '
                            '"domain": "0000", '
                            '"function": "0"}',
             'in_use': False}
        ]
        attribute_list = [
            {'key': 'rc', 'value': 'CUSTOM_SSD'},
            {'key': 'trait0', 'value': 'CUSTOM_SSD_INSPUR'},
            {'key': 'trait1', 'value': 'CUSTOM_SSD_PRODUCT_ID_1001'},
        ]
        expected = {
            'vendor': '1bd4',
            'type': 'SSD',
            'std_board_info':
                {"controller": "Non-Volatile memory controller",
                 "product_id": "1001"},
            'vendor_board_info': {"vendor_info": "ssd_vb_info"},
            'deployable_list':
                [
                    {
                        'num_accelerators': 1,
                        'driver_name': 'INSPUR',
                        'name': 'host-192-168-32-195_0000:db:00.0',
                        'attach_handle_list': attach_handle_list,
                        'attribute_list': attribute_list
                    },
                ],
            'controlpath_id': {'cpid_info': '{"bus": "db", '
                                            '"device": "00", '
                                            '"domain": "0000", '
                                            '"function": "0"}',
                               'cpid_type': 'PCI'}
        }
        ssd_obj = ssd_list[0]
        ssd_dict = ssd_obj.as_dict()
        ssd_dep_list = ssd_dict['deployable_list']
        ssd_attach_handle_list = \
            ssd_dep_list[0].as_dict()['attach_handle_list']
        ssd_attribute_list = \
            ssd_dep_list[0].as_dict()['attribute_list']
        attri_obj_data = []
        [attri_obj_data.append(attr.as_dict()) for attr in ssd_attribute_list]
        attribute_actual_data = sorted(attri_obj_data, key=lambda i: i['key'])
        self.assertEqual(expected['vendor'], ssd_dict['vendor'])
        self.assertEqual(expected['controlpath_id'],
                         ssd_dict['controlpath_id'])
        self.assertEqual(expected['std_board_info'],
                         jsonutils.loads(ssd_dict['std_board_info']))
        self.assertEqual(expected['vendor_board_info'],
                         jsonutils.loads(ssd_dict['vendor_board_info']))
        self.assertEqual(expected['deployable_list'][0]['num_accelerators'],
                         ssd_dep_list[0].as_dict()['num_accelerators'])
        self.assertEqual(expected['deployable_list'][0]['name'],
                         ssd_dep_list[0].as_dict()['name'])
        self.assertEqual(expected['deployable_list'][0]['driver_name'],
                         ssd_dep_list[0].as_dict()['driver_name'])
        self.assertEqual(attach_handle_list[0],
                         ssd_attach_handle_list[0].as_dict())
        self.assertEqual(attribute_list, attribute_actual_data)

    @mock.patch('cyborg.accelerator.drivers.ssd.utils.lspci_privileged')
    def test_discover_with_base_ssd_driver(self, mock_devices_for_vendor):
        mock_devices_for_vendor.return_value = self.p.stdout.readlines()
        with self.assertLogs(None, level='INFO') as cm:
            d = SSDDriver.create()
            ssd_list = d.discover()
        self.assertEqual(cm.output,
                         ['INFO:cyborg.accelerator.drivers.ssd.base:The '
                          'method "discover" is called in generic.SSDDriver'])
        self.assertEqual(2, len(ssd_list))
        self.assertEqual("1bd4", ssd_list[1].as_dict()['vendor'])
